
from .dframe import DFrame, Databunch
from .kagglereader import Kaggle
from torchvision.datasets import ImageFolder
from torchvision.transforms import transforms
from torch.utils.data import random_split
import torch
import pandas as pd
import numpy as np
import torch.nn.functional as F
import matplotlib.pyplot as plt
from PIL import Image, ImageStat
import random
from pathlib import Path

def LoadImage(fp):
    return Image.open(fp[0])

def LoadImageFast(fp):
    import pyvips
    image = pyvips.Image.new_from_file(fp[0], access="sequential")
    image = image.colourspace("srgb")
    mem_img = image.write_to_memory()
    return np.frombuffer(mem_img, dtype=np.uint8).reshape(image.height, image.width, 3)

def _subplots(rows, cols, imgsize=4, figsize=None, title=None, **kwargs):
    "Like `plt.subplots` but with consistent axs shape, `kwargs` passed to `fig.suptitle` with `title`"
    if figsize is None:
        figsize = (imgsize*cols, imgsize*rows)
    fig, axs = plt.subplots(rows,cols,figsize=figsize)
    if rows==cols==1:
        axs = [[axs]]
    elif (rows==1 and cols!=1) or (cols==1 and rows!=1):
        axs = [axs]
    if title is not None:
        fig.suptitle(title, **kwargs)
    return np.array(axs)

def _show_batch(ds, rows=3, imgsize=(20,20), figsize=(10,10), classes=None, normalized_mean=None, normalized_std=None):
    inv_normalizer = lambda x:x
    if normalized_std is not None and normalized_mean is not None:
            inv_normalizer = transforms.Normalize(mean=tuple(-m/s for m, s in zip(normalized_mean, normalized_std)), std=tuple(1/s for s in normalized_std))

    axs = _subplots(rows, rows, imgsize=imgsize, figsize=figsize).flatten()

    for i, ax in enumerate(axs):
        img, y = ds[random.randrange(0, len(ds))]
        img = transforms.Resize([100,100], antialias=True)(img)
        img = inv_normalizer(img)
        try:
            img = transforms.ToTensor()(img)
        except: pass
        try:
            img = img.numpy()
#                 if img.shape[0] == 1:
#                     img = img.squeeze(axis=0)
#                 elif img.shape[0] == 3:
#                     img = img.permute(1, 2, 0)
        except: pass
        if img.shape[0] == 1:
            img = img.squeeze(axis=0)
        elif img.shape[0] == 3:
            img = np.transpose(img, (1, 2, 0))
        img = np.clip(img, 0, 1)    
        ax.imshow(img)
        try:
            y = classes[int(y)]
        except:
            pass
        ax.set_title(f'y={y}')
    for ax in axs.flatten()[i:]:
        ax.axis('off')
    plt.tight_layout()
    plt.show()

class ImageDatabunch(Databunch):
    def __init__(self, train_ds, test_ds=None, valid_ds=None, batch_size=32, 
                 test_batch_size=None, num_workers=2, shuffle=True, pin_memory=False, 
                 collate=None, balance=None, normalized_mean=None, normalized_std=None, classes=None):
        if test_batch_size is None:
            test_batch_size = batch_size
        try:
            self.classes = classes or train_ds.classes
        except:
            try:
                self.classes = classes or train_ds.dataset.classes
            except:
                self.classes = None
        try:
            self.normalized_mean = normalized_mean or train_ds.normalized_mean
            self.normalized_std = normalized_std or train_ds.normalized_std
        except:
            try:
                self.normalized_mean = normalized_mean or train_ds.dataset.normalized_mean
                self.normalized_std = normalized_std or train_ds.dataset.normalized_std
            except: pass
        super().__init__(None, train_ds, test_ds=test_ds, valid_ds=valid_ds, batch_size=batch_size, 
                         test_batch_size=test_batch_size, num_workers=num_workers, shuffle=shuffle, 
                         pin_memory=pin_memory, collate=collate, balance=balance)

    @classmethod
    def from_train_test_ds(cls, train_ds, test_ds, valid_perc=0.2, 
                           batch_size=32, test_batch_size=None, 
                           num_workers=2, shuffle=True, pin_memory=False, collate=None,
                           balance=None,
                           classes=None, normalized_mean=None, normalized_std=None):
        valid_length = int(valid_perc * len(train_ds))
        train_length = len(train_ds) - valid_length
        if shuffle:
            train_ds, valid_ds = random_split(train_ds, [train_length, valid_length])
        else:
            train_ds = train_ds[:train_length]
            valid_ds = train_ds[train_length:]
        return cls(train_ds, test_ds, valid_ds, batch_size=batch_size, 
                   test_batch_size=test_batch_size, num_workers=num_workers, shuffle=shuffle, 
                   pin_memory=pin_memory, collate=collate, balance=balance, classes=classes,
                   normalized_mean=normalized_mean, normalized_std=normalized_std)
              
    def show_batch(self, rows=3, imgsize=(20,20), figsize=(10,10)):
        try:
            _show_batch( self.train_ds, rows=rows, imgsize=imgsize, figsize=figsize, classes=self.classes,
                       normalized_mean=self.normalized_mean, normalized_std=self.normalized_std)
        except:
            _show_batch( self.train_ds, rows=rows, imgsize=imgsize, figsize=figsize, classes=self.classes)
            

class ImageDFrame(DFrame):
    _metadata = DFrame._metadata + ['_pt_dataset_transforms', '_pt_normalize', '_pt_normalized_mean', '_pt_normalized_std',
                                    '_pt_classes', '_pt_class_to_idx']

    _internal_names = DFrame._internal_names + ['_pt__locked_normalized_mean', '_pt__locked_normalized_std']
    
    _internal_names_set = set( _internal_names )
    
    def __init__(self, data, *args, classes=None, **kwargs):
        super().__init__(data, *args, **kwargs)
        self._pt_dtype = str
        self._pt__locked_normalized_mean = None
        self._pt__locked_normalized_std = None
        self.normalize()
        self._pt_classes = classes
        
    @classmethod
    def read_from_kaggle(cls, dataset, filename=None, shared=True, force=False, **kwargs):
        k = Kaggle(dataset, shared=shared)
        if force:
            k.remove_user()
        folder = k.file(filename)
        subfolders = len([ 1 for file in folder.glob('**/*') if file.is_dir() ])
        if subfolders > 1:
            return cls.from_image_folder(folder)
        else:
            return cls.from_image_files(folder, **kwargs)

    def to_databunch(self, datasetclass=None, batch_size=32, test_batch_size=None, 
                     num_workers=2, shuffle=True, pin_memory=False, 
                     balance=None, collate=None):
        train, test, valid = self.to_datasets(datasetclass=datasetclass)
        if valid is None:
            valid = test
        return ImageDatabunch(train, test, valid, 
                         batch_size=batch_size, test_batch_size=test_batch_size,
                         num_workers=num_workers, shuffle=shuffle, 
                         pin_memory=pin_memory, balance=balance, collate=collate, classes=self.classes,
                         normalized_mean=self.normalized_mean, normalized_std=self.normalized_std)  
    
    @property
    def _constructor(self):
        return ImageDFrame

    @property
    def classes(self):
        return self._pt_classes
    
    @property
    def class_to_idx(self):
        return self._pt_class_to_idx
    
    @classes.setter
    def classes(self, value):
        if value is not None:
            self._pt_classes = value
            self._pt_class_to_idx = { c:i for i, c in enumerate(value) }
    
    def _dataset_pre_transforms(self):
        return [ LoadImage ]
    
    def _dataset_post_transforms(self):
        return [ transforms.ToTensor() ]
    
    def _train_transformation_parameters(self, train_dset):
        """
        This will be called from DFrame when the datasets are consructed on the train dataset.
        """
        if self._pt_normalize:
            if self._pt_normalized_mean is None or self._pt_normalized_std is None:
                n = 0
                sample = range(len(train_set)) if len(train_dset) < 100 else random.sample(range(len(train_dset)), min(100, len(train_dset)))
                dset = train_dset.to_dataset()
                for i in sample:
                    x, y = dset[i]
                    try:
                        channels
                    except:
                        channels = len(x)
                        x1 = [0.0 for i in range(channels)]
                        x2 = [0.0 for i in range(channels)]

                    for c in range(channels):
                        x1[c] += x[c].view(-1).sum()
                        x2[c] += (x[c].view(-1) ** 2).sum()
                    n = n + x[0].shape[0] * x[0].shape[1]
                sd = np.zeros(channels)
                mean = np.zeros(channels)
                for c in range(channels):
                    sd[c] = np.sqrt(((n * x2[c]) - (x1[c] * x1[c])).numpy() / (n * (n - 1)))
                    mean[c] = (x1[c] * x1[c]).numpy() / (n * (n - 1))
                self._pt__locked_normalized_mean = torch.tensor(mean) 
                self._pt__locked_normalized_std = torch.tensor(sd)
                
            else:
                self._pt__locked_normalized_mean = self._pt_normalized_mean
                self._pt__locked_normalized_std = self._pt_normalized_std
            return True

    def normalize(self, do=True, normalized_mean=None, normalized_std=None):
        """
        Normalize the images. You can either supply normalize with a precalculated mean and standard
        deviation per channel, or pass True to automatically calculate these parameters on the 
        training set. This will require an additional pass over the training set.
        
        Arguments:
            do: bool (True) - if True images are normalized
            
            normalized_mean: (float)
                set of precalculated means for the dataset. 
                The set size should be the same as the number of channels in the images.
                
            normalized_std: (float)
                set of precalculated standard deviations for the dataset
                The set size should be the same as the number of channels in the images.
        """
        self._pt_normalize = do
        self._pt_normalized_mean = torch.tensor(normalized_mean) if normalized_mean is not None else None
        self._pt_normalized_std = torch.tensor(normalized_std) if normalized_std is not None else None
        return self

    @property
    def normalized_mean(self):
        return self._pt__locked_normalized_mean    
        
    @normalized_mean.setter
    def normalized_mean(self, value):
        self._pt__locked_normalized_mean = value
    
    @property
    def normalized_std(self):
        return self._pt__locked_normalized_std
    
    @normalized_std.setter
    def normalized_std(self, value):
        self._pt__locked_normalized_std = value
        
    def _dataset_transforms(self, pre=True, train=True, standard=True, post=True, normalize=True):
        t = super()._dataset_transforms(pre=pre, train=train, standard=standard, post=post)
        if normalize and self._pt__locked_normalized_mean is not None and self._pt__locked_normalized_std is not None:
            t.append(transforms.Normalize(mean=self._pt__locked_normalized_mean, 
                                          std=self._pt__locked_normalized_std))
        return t
    
    def train_images_ds(self):
        """
        returns a version of the train DataSet, for which normalization and post_transforms are turned off
        in other words, this will return the images just before they are converted to tensors. 
        """
        return self.dtype(False)._dset_indices(self._train_indices, self._dataset_transforms(post=False, normalize=False)).to_dataset()
    
    def show_batch(self, rows=3, imgsize=(20,20), figsize=(10,10)):
        """
        Shows a sample of rows*rows images from the training set.
        """
        _show_batch( self.train_images_ds(), rows=rows, imgsize=imgsize, figsize=figsize, classes=self.classes)
    
    @classmethod
    def from_binary_folder(cls, folder, **kwargs):
        """
        Construct an ImageDFrame from the filelist that is obtained from TorchVision's ImageFolder,
        in other words, the subfolders are the class labels assigned to the files they contain.
        
        from_binary_folder assumes that we will attempt to use the dataset with a BCELoss functions
        and put the target variable as float32 in a column (2D) vector 
        """
        try:
            folder.samples
        except:
            folder = ImageFolder(root=folder)
        r = ImageDFrame(folder.samples, columns=['filename', 'target'], classes=folder.classes)
        r.target = r.target.astype(np.int64)
        return r
    
    @classmethod
    def from_image_folder(cls, folder):
        """
        Construct an ImageDFrame from the filelist that is obtained from TorchVision's ImageFolder,
        in other words, the subfolders are the class labels assigned to the files they contain.

        from_image_folder assumes the data is used with a CrossEntropyLoss function
        and puts the target variable as long in a row (1D) vector 
        """
        try:
            folder.samples
        except:
            folder = ImageFolder(root=folder)
        r = ImageDFrame(folder.samples, columns=['filename', 'target'], classes=folder.classes)
        r.target = r.target.astype(np.int64)
        return r

    @classmethod
    def from_image_files(cls, folder, ext=None, omit=None, delimiter='.', targetindex=0):
        """
        Construct an ImageDFrame from a folder with files in which the
        part of the filename indicates the class label, e.g. horse.1.jpg,
        cow.2.jpg.
        
        The label is found using the delimiter and targetindex. In the given example
        a delimiter='.' and targetindex=0 would use 'horse' and 'cow' as labels and
        targetindex=-2 would use 1 and 2. 
        
        Three different target labels are supported: numeric, text and None: 
            targetindex=None indicates there is no target label and the filename
            is also used as the targetlabel.

            Otherwise, an attempt is made to convert the label into a float32, 
            since this is commonly used for PyTorch. If that succeeds
            a numeric label is used. 

            If float conversion fails, a categorical label is assumed and 
            DFrame.categoryy() is set accordingly, meaning that the DataFrame will show
            a String label, but during data preparation the target values are converted 
            into unique numbers.

        Args:
            folder: str or Path
                folder containing the files
            ext: str (None)
                extension of the files that are included as images
            omit: [ str ] (None)
                omit all files that end with any of the give strings
            delimiter: str ('.')
                delimiter used to obtain the class label from the filename.
                By default a '.' is used, in which case horse.1.jpg will use the
                name 'horse' as its class label.
            targetindex: int/None (0)
                To obtain the target, the filenames are split by the given delimiter
                and then targetindex indicates the position of the targetlabel, 
                e.g. -2 for 'horse.1.jpg' will use '1', -1 would use 'jpg' and
                0 would use 'horse'. 
                
        """
        if type(folder) == str:
            folder = Path(folder)
        if ext:
            files = list(folder.glob(f'*.{ext}'))
        else: 
            files = list(folder.glob('*'))
        if omit is not None:
            for o in omit:
                files = [ f for f in files if not f.endswith(o) ]
        files = [ f for f in files if '/.' not in str(f)[len(str(folder)):] ]
        if targetindex is None:
            files = [ (str(f), str(f)) for f in files ]
            return ImageDFrame(files, columns=['filename', 'target'])
        
        target = [ ('/' + str(f)).split('/')[-1] for f in files ]
        target = [ t.split(delimiter)[targetindex] for t in target ]
        try:
            target = [ float(t) for t in target ]
            converted_to_float = True
        except:
            converted_to_float = False
        files = [ (str(f), t) for f, t in zip(files, target) ]
        r = ImageDFrame(files, columns=['filename', 'target'])
        if converted_to_float:
            if r.target.min() == 0.0 and r.target.max() == 1.0:
                r.target = r.target.astype(np.float32)
            else:
                r.target = r.target.astype(np.int64)
        else:
            r = r.category('target')
        return r



import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler, PolynomialFeatures, OneHotEncoder, FunctionTransformer
from sklearn.utils import resample
from sklearn.model_selection import KFold, train_test_split, StratifiedKFold, StratifiedShuffleSplit
from iterstrat.ml_stratifiers import MultilabelStratifiedKFold, MultilabelStratifiedShuffleSplit
import matplotlib.pyplot as plt
import copy
import math
import time
import random
import warnings
import linecache
from ..evaluate.evaluate import EvaluatorSK
from .helper import read_from_package, read_from_function
from .kagglereader import Kaggle
from .databunch import Databunch
from .dset import DSet
from pandas.core.groupby.generic import DataFrameGroupBy, SeriesGroupBy
from collections import defaultdict
from contextlib import contextmanager
import functools
from pandas.util import hash_pandas_object
import hashlib
from pandas.util._exceptions import find_stack_level
import warnings

pd.options.mode.chained_assignment = None

def to_numpy(arr):
    try:
        return arr.data.cpu().numpy()
    except: pass
    try:
        return arr.to_numpy()
    except: pass
    return arr

class show_warning:
    def __enter__(self):
        self.warning = warnings.catch_warnings(record=True)
        self.w = self.warning.__enter__()
        warnings.filterwarnings('error')
        warnings.simplefilter('default')
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        for wi in self.w:
            if wi.line is None:
                print(wi.filename)
                wi.line = linecache.getline(wi.filename, wi.lineno)
        print(f'line number {wi.lineno}  line {wi.line}') 
        self.warning.__exit__(exc_type, exc_value, exc_traceback)
        
class _DFrame:
    _config   = ['_pt_scale_columns', '_pt_scale_omit_interval', '_pt_scalertype', '_pt_columny', '_pt_columnx', 
                 '_pt_vectory', '_pt_bias', '_pt_polynomials', '_pt_dtype', '_pt_category', '_pt_category_sort', 
                 '_pt_dummies', '_pt_evaluator', '_pt_transform',
                 '_pt_dataset_transforms', '_pt_dataset_train_transforms' ]
    
    _config_indices = [
                 '_pt_sequence_window', '_pt_sequence_shift_y', 
                 '_pt_split_shuffle', '_pt_split_stratify', '_pt_split_stratify_test', 
                 '_pt_folds_stratify', '_pt_folds_random_state', 
                 '_pt_split_random_state', '_pt_folds_shuffle', 
                 '_pt_valid_size', '_pt_test_size', '_pt_balance', '_pt_filterna', 
                 '_pt_train_test_indices', '_pt_test_indices', '_pt_folds', '_pt_fold',
                 '_cached_index', '_pt_folds_mode' ]
        
    _config_fold = [ '_pt_fold' ]
    
    _cached =  [ #'_cached_changed', #'_cached_changed_fold', 
                 '_cached_train', '_cached_valid', '_cached_test', '_cached_raw_train',
                 '_cached_scalerx', '_cached_scalery',
                 '_cached_categoryx', '_cached_categoryy',
                 '_cached_columntransformerx', '_cached_columntransformery',
                 '_cached_dummiesx', '_cached_dummiesy']
                
    _cached_indices = [ '_cached_train_indices', '_cached_valid_indices', '_cached_test_indices',
                        '_cached_indices_before_testsplit', '_cached_indices_after_testsplit' ]
    
    _cached_fold = [ '_cached_folds' ]

    _metadata = _config + _config_indices + _config_fold + _cached + _cached_indices + _cached_fold
    _config_set = set( _config )
    _config_indices_set = set( _config_indices )
    _config_fold_set = set( _config_fold )
    
    _internal_names = pd.DataFrame._internal_names + _metadata
    
    _internal_names_set = set( _internal_names )
    
    def __init__(self, data, **kwargs):
        for m in self._config:
            setattr(self, m, None)
        for m in self._config_indices:
            setattr(self, m, None)
        self._index_changed()
        for m in self._metadata:
            try:
                setattr(self, m, getattr(data, m))
            except: pass
    
    @classmethod
    def stop_warnings(cls):
        pd.options.mode.chained_assignment = None
    
    def _copy_meta(self, r):
        for c in self._metadata:
            try:
                setattr(r, c, getattr(self, c))
            except: pass
        return r

    def _copy_non_cached_meta(self, r):
        for c in self._config_set:
            try:
                setattr(r, c, getattr(self, c))
            except: pass
        for c in self._config_fold_set:
            try:
                setattr(r, c, getattr(self, c))
            except: pass
        for c in self._config_indices_set:
            try:
                setattr(r, c, getattr(self, c))
            except: pass
        return r

    def _dframe(self, data):
        return self._copy_meta( DFrame(data) )

    def _dset(self, data, transforms=None):
        return DSet.from_dframe(data, self, transforms)
        
    @property
    def _columny(self):
        try:
            if len(self._pt_columny) > 0:
                return self._pt_columny
        except: pass
        return [ self.columns[-1] ]
        
    def dtype(self, dtype):
        """
        Lazily executed change of dtype to the data.
        """
        r = self.copy(deep=False)
        r._change('_pt_dtype', dtype)
        return r
        
    @property
    def _columnx(self):
        if self._pt_columnx is None:
            return [ c for c in self.columns if c not in self._columny ]
        return [ c for c in self._pt_columnx if c not in self._columny ]
   
    @property
    def _all_columns(self):
        return list(set(self._columnx).union(set(self._columny)))

    @property
    def _columnsx_scale_indices(self):
        if self._pt_polynomials is not None and self._pt_scale_columns is not None:
            X = self.raw_train._x_polynomials
            return [ i for i in range(X.shape[1]) if (X[:,i].min() < self._pt_scale_omit_interval[0] or X[:,i].max() > self._pt_scale_omit_interval[1]) ]
        columnx = self._columnx
        cat = set(self._pt_category) if type(self._pt_category) == tuple else []
        if self._pt_scale_columns == True or self._pt_scale_columns == 'x_only':
            r = [ c for c in columnx if c not in cat]
        elif self._pt_scale_columns == False or self._pt_scale_columns is None or len(self._pt_scale_columns) == 0:
            r = []
        else:
            r = [ c for c in columnx if c in self._pt_scale_columns and c not in cat ]
        if len(r) > 0:
            X = self.raw_train._x_polynomials
            r = [ columnx.index(c) for i, c in enumerate(columnx) if c in r and ((X[:,i].min() < self._pt_scale_omit_interval[0] or X[:,i].max() > self._pt_scale_omit_interval[1])) ]
        return r
        
    @property
    def _columnsy_scale_indices(self):
        columny = self._columny
        cat = set(self._pt_category) if type(self._pt_category) == tuple else []
        if self._pt_scale_columns == True:
            y = self.raw_train._y_numpy
            r = [ c for i, c in enumerate(columny) if c not in cat and (y[:,i].min() < self._pt_scale_omit_interval[0] or y[:,i].max() > self._pt_scale_omit_interval[1]) ]
        elif self._pt_scale_columns == False or self._pt_scale_columns is None or len(self._pt_scale_columns) == 0:
            r = []
        else:
            r = [ c for c in columny if c in self._pt_scale_columns and c not in cat ]
        return [ columny.index(c) for c in r ]

    @staticmethod
    def _create_scaler(scalertype, column):
        scaler = scalertype()
        scaler.fit(column)
        return scaler
    
    @property
    def _scalerx(self):
        if self._unchanged() and self._cached_scalerx is not None:
            return self._cached_scalerx
        
        X = self.raw_train._x_polynomials
        r = [ None ] * X.shape[1]
        for i in self._columnsx_scale_indices:
            r[i] = self._create_scaler(self._pt_scalertype, X[:, i:i+1])
        self._change('_cached_scalerx', r)
        return r
     
    @_scalerx.setter
    def _scalerx(self, value):
        self._change('_cached_scalerx', value)
        
    @property
    def _scalery(self):
        if self._unchanged() and self._cached_scalery is not None:
            return self._cached_scalery
        
        y = self.raw_train._y_numpy
        r = [ None ] * y.shape[1]
        for i in self._columnsy_scale_indices:
            r[i] = self._create_scaler(self._pt_scalertype, y[:, i:i+1])
        self._change('_cached_scalery', r)
        return r
    
    @_scalery.setter
    def _scalery(self, value):
        self._change('_cached_scalery', value)
        
    def _create_category(self, column):
        sort = self._pt_category_sort
        class Category:
            def fit(self, X):
                s = X.unique()
                if sort:
                    s = sorted(s)
                self.dict = defaultdict(lambda:0, { v:(i) for i, v in enumerate(s) })
                self.inverse_dict = { (i):v for i, v in enumerate(s) }
                self.inverse_dict[0] = np.NaN
            
            def transform(self, X):
                return X.map(self.dict)
            
            def inverse_transform(self, X):
                return X.map(self.inverse_dict)
            
        if column not in self._pt_category:
            return None
        
        c = Category()
        c.fit(self.raw_train[column])
        return c
    
    def _categoryx(self):
        if self._unchanged() and self._cached_categoryx is not None:
            return self._cached_categoryx
        
        if self._pt_category is None or len(self._pt_category) == 0:
            return None
        r = [ self._create_category(c) for c in self._columnx ]
        self._change('_cached_categoryx', r)
        return self._cached_categoryx
    
    def _categoryy(self):
        if self._unchanged() and self._cached_categoryy is not None:
            return self._cached_categoryy
        
        if self._pt_category is None or len(self._pt_category) == 0:
            return None
        r = [ self._create_category(c) for c in self._columny ] 
        self._change('_cached_categoryy', r)
        return r

    def _columntransformerx(self):
        if self._unchanged() and self._cached_columntransformerx is not None:
            return self._cached_columntransformerx
        
        if self._pt_transform is None or len(self._pt_transform) == 0:
            return None
        r = []
        for c in self._columnx:
            try:
                t = self._pt_transform[c]
                t.fit(self.raw_train[c])
                r.append(t)
            except:
                r.append(None)
                
        self._change('_cached_columntransformerx', r)
        return self._cached_columntransformerx
        
    def _columntransformery(self):
        if self._unchanged() and self._cached_columntransformery is not None:
            return self._cached_columntransformery
        
        if self._pt_transform is None or len(self._pt_transform) == 0:
            return None
        r = []
        for c in self._columny:
            try:
                t = self._pt_transform[c]
                t.fit(self.raw_train[c])
                r.append(t)
            except:
                r.append(None)
                
        self._change('_cached_columntransformery', r)
        return self._cached_columntransformery
    
    def _create_dummies(self, column):    
        if column not in self._pt_dummies:
            return None
        
        c = OneHotEncoder(handle_unknown='ignore')
        c.fit(self.raw_train[[column]])
        return c
    
    def _dummiesx(self):
        if self._unchanged() and self._cached_dummiesx is not None:
            return self._cached_dummiesx
        
        if self._pt_dummies is None or len(self._pt_dummies) == 0:
            r = [ None ] * len(self._columnx)
        else:
            r = [ self._create_dummies(c) for c in self._columnx ]
        self._change('_cached_dummiesx', r)
        return r
    
    def _dummiesy(self):
        if self._unchanged() and self._cached_dummiesy is not None:
            return self._cached_dummiesy
        
        if self._pt_dummies is None or len(self._pt_dummies) == 0:
            r = [ None ] * len(self._columny)
        else:
            r = [ self._create_dummies(c) for c in self._columny ]
        self._change('_cached_dummiesy', r)
        return r
        
    @property
    def _shift_y(self):
        if self._pt_sequence_shift_y is not None:
            return self._pt_sequence_shift_y
        else:
            return 0
    
    @property
    def _sequence_window(self):
        try:
            if self._pt_sequence_window is not None:
                return self._pt_sequence_window
        except:pass
        return 1
    
    @property
    def _sequence_index_y(self):
        return self._sequence_window+self._shift_y-1
        
    @property
    def _indices_unshuffled(self):
        if self._pt_train_test_indices is not None:
            return np.intersect1d(self.index.values, self._pt_train_test_indices)
        return self.index.values

    @property
    def _indices_before_testsplit(self):
        if self._unchanged() and self._cached_indices_before_testsplit is not None:
            return self._cached_indices_before_testsplit
        
        test_indices = set(self.fixed_test_indices)
        if len(test_indices) > 0:
            r = np.array([ i for i in self._indices_unshuffled if i not in test_indices ])
        else:
            r = self._indices_unshuffled
        self._change('_cached_indices_before_testsplit', r)
        return r
           
    @property
    def _indices_after_testsplit(self):
        if self._unchanged() and self._cached_indices_after_testsplit is not None:
            return self._cached_indices_after_testsplit
        
        test_indices = set(self._test_indices)
        r = np.array([ i for i in self._indices_before_testsplit if i not in test_indices ])
        self._change('_cached_indices_after_testsplit', r)
        return r

    @property
    def _test_size(self):
        return self._pt_test_size or 0
    
    @property
    def _valid_size(self):
        return self._pt_valid_size or 0
    
    @property
    def _shuffle(self):
        return ((self._pt_split_shuffle is None and \
                (self._test_size > 0 or self._valid_size > 0)) \
                or self._pt_split_shuffle) and self._pt_sequence_window is None
    
    def _stratifyable_columns(self, indices, columns, bins):
        if columns == True:
            columns = self._all_columns
        df = self.loc[indices, columns]
        if bins > 1:
            bins = len(df) // bins
        else:
            if bins >= 0.5:
                bins = 1 - bins
            bins = math.ceil(1 / bins)
            bins = len(df) // bins
        for c in df.columns:
            if pd.api.types.is_float_dtype(df[c]):
                df[c] = pd.qcut(df[c] + np.random.random(len(df))/1000, bins, labels=False)
        return df
    
    @property
    def fixed_test_indices(self):
        if self._pt_test_indices is not None:
            r = np.intersect1d(self.index.values, self._pt_test_indices)
            if len(r) < len(self._pt_test_indices):
                warnings.warn("Test rows were lost because of a previous operation.",
                    RuntimeWarning,
                )
                self._pt_test_indices = r
            return self._pt_test_indices
        return []
        
    @property
    def _test_indices(self):
        if self._pt_test_indices is not None:
            return self.fixed_test_indices
        if self._unchanged() and self._cached_test_indices is not None:
            return self._cached_test_indices
        if self._pt_folds is not None and (self._pt_folds_mode == 'both' or
                                           self._pt_folds_mode == 'test'):
            r = self._test_fold
        elif self._test_size > 0:
            if self._pt_split_stratify_test is None:
                if self._shuffle:
                    _, r = train_test_split(self._indices_before_testsplit, test_size=self._test_size, random_state=self._pt_split_random_state)
                    r = sorted(r)
                else:
                    test_size = int(self._test_size * len(self._indices_before_testsplit))
                    r = self._indices_before_testsplit[-test_size:]
            else:
                if self._pt_split_stratify_test == True or len(self._pt_split_stratify_test) > 1:
                    splitter = MultilabelStratifiedShuffleSplit(n_splits=1, 
                                                       random_state=self._pt_split_random_state, 
                                                       test_size=self._test_size)
                else:
                    splitter = StratifiedShuffleSplit(n_splits=1, 
                                                       random_state=self._pt_split_random_state, 
                                                       test_size=self._test_size) 
                target = self._stratifyable_columns(self._indices_before_testsplit, 
                                                    self._pt_split_stratify_test, self._test_size / len(self))
                _, r = next(splitter.split(target, target))
                r = sorted(r)
        else:
            r = []  
        self._change('_cached_test_indices', r)
        return r
    
    @property
    def _train_indices_unbalanced(self):
        return sorted(set(self._indices_after_testsplit) - set(self._valid_indices))

    def _pseudo_choose(self, indices, items):
        if self._pt_split_random_state is not None:
            random.seed(self._pt_split_random_state)
        r = np.array(random.sample(sorted(indices), items % len(indices)))
        r = np.hstack([indices for i in range(items // len(indices))] + [r])
        return r

    @property
    def _train_indices(self):
        if self._unchanged() and self._cached_train_indices is not None:
            return self._cached_train_indices
        r = self._train_indices_unbalanced
        if self._pt_balance is not None:
            y = self.loc[r, self._columny]
            classes = np.unique(y)
            classindices = {c:np.where(y==c)[0] for c in classes}
            classlengths = {c:len(indices) for c, indices in classindices.items()}
            if self._pt_balance == True: # equal classes
                n = max(classlengths.values())
                mask = np.hstack([self._pseudo_choose(classindices[c], n) for c in classes])
            else:                        # use given weights
                weights = self._pt_balance
                n = max([ int(math.ceil(classlengths[c] / w)) for c, w in weights.items() ])
                mask = np.hstack([self._pseudo_choose(classindices[c], round(n*weights[c])) for c in classes])
            r = np.array(r)[ mask.astype(int) ]
        self._change('_cached_train_indices', r)
        return r

    @property
    def _valid_indices(self):
        if self._unchanged() and self._cached_valid_indices is not None:
            return self._cached_valid_indices
        if self._pt_folds is not None:
            r = self._valid_fold
        elif self._valid_size > 0:
            valid_size = self._valid_size / (1 - self._test_size)
            if valid_size > 0:
                if self._pt_split_stratify is None:
                    if self._shuffle:
                        _, r = train_test_split(self._indices_after_testsplit, test_size=valid_size, random_state=self._pt_split_random_state)
                        r = sorted(r)
                    else:
                        valid_size = int(valid_size * len(self._indices_before_testsplit))
                        r = self._indices_after_testsplit[-valid_size:]                        
                else:
                    if self._pt_split_stratify == True or len(self._pt_split_stratify) > 1:
                        splitter = MultilabelStratifiedShuffleSplit(n_splits=1, 
                                                           random_state=self._pt_split_random_state, 
                                                           test_size=valid_size)
                    else:
                        splitter = StratifiedShuffleSplit(n_splits=1, 
                                                           random_state=self._pt_split_random_state, 
                                                           test_size=valid_size) 
                    target = self._stratifyable_columns(self._indices_after_testsplit, self._pt_split_stratify, self._valid_size / (len(self)-self._test_size))
                    _, r = next(splitter.split(target, target))
                    r = sorted(r)
        else:
            r = []
        self._change('_cached_valid_indices', r)
        return r

    @property
    def _folds(self):
        if self._unchanged() and self._cached_folds is not None:
            return self._cached_folds
        r = []
        if self._test_size > 0 and self._pt_folds_mode == 'valid':
            indices = self._indices_after_testsplit
        else:
            indices = self._indices_before_testsplit
        if self._pt_folds_stratify is None:
            target = self.loc[indices]
            splitter = KFold(n_splits = self._pt_folds, shuffle=self._pt_folds_shuffle, random_state=self._pt_folds_random_state)
            for train_indices, test_indices in splitter.split(target, target):
                r.append(sorted(indices[test_indices]))
        else:
            if self._pt_folds_stratify != True and len(self._pt_folds_stratify) > 1:
                splitter = MultilabelStratifiedKFold(n_splits = self._pt_folds,
                                    shuffle=True,
                                    random_state=self._pt_folds_random_state)
            else:
                splitter = StratifiedKFold(n_splits = self._pt_folds,
                                    shuffle=True,
                                    random_state=self._pt_folds_random_state)
            target = self._stratifyable_columns(indices, self._pt_folds_stratify, self._pt_folds)
            for train_indices, test_indices in splitter.split(target, target):
                r.append(sorted(indices[test_indices]))
        self._change('_cached_folds', r)
        return r
        
    @property
    def _valid_fold(self):
        """
        the current test fold, determined by df.fold(i) + 1
        """
        assert self._pt_fold is not None, 'Apply folds() first'
        assert self._pt_folds_mode == 'both' or self._pt_folds_mode == 'valid', 'There is no valid fold'
        if self._pt_folds_mode == 'both':
            valid_fold = (self._pt_fold + 1 + (self._pt_fold // self._pt_folds % (self._pt_folds - 1))) % self._pt_folds
        else:
            valid_fold = self._pt_fold
        return self._folds[valid_fold]
    
    @property
    def _test_fold(self):
        """
        The current valid fold, determined by df.fold(i)
        """
        assert self._pt_fold is not None, 'Apply folds() first'
        if self._pt_folds_mode == 'valid':
            return self._test_indices
        return self._folds[self._pt_fold]
    
    def fold(self, i):
        """
        Utilize n-fold cross validation by first calling `folds(n)` and then calling `fold(i)` to obtain 
        a DFrame in which cross validation is set up using the fold (i) as the validation set, 
        another fold as the test set and the remainder as the training set.
        
        A first call will trigger rows to be assigned to the train, valid and test part, which are stored
        in place to reproduce the exact same split for consecutive calls. The call will however return
        a copy of the DataFrame in which the requested fold is selected.

        Arguments:
            i: int
                the fold to use as the validation set
        
        Returns: copy of the PipeTorch DataFrame
            In this copy, the train, valid and test sets are shifted to fold n
        """
        self._folds
        r = self.copy(deep=False)
        
        i = i % r._pt_folds
        r._change('_pt_fold', i)
        return r
    
    def iterfolds(self):
        """
        Iterate over the folds for n-fold cross validation. 
        
        A first call will trigger rows to be assigned to the train, valid and test part, which are stored
        in place to reproduce the exact same split for consecutive calls.
        
        Yields:
            train, valid (DSet)
        """
        for i in range(self._pt_folds):
            yield self.fold(i)
    
    def df_to_dataset(self, df, datasetclass=None):
        """
        Converts the given df to a DataSet using the pipeline of this DFrame.
        
        Arguments:
            df: DataFrame or DFrame
                to convert into a DataSet
        
        returns: Converts the given df to a DataSet.
        """
        return self.df_to_dset(df).to_dataset(datasetclass)
        
    def to_datasets(self, datasetclass=None):
        """
        Prepares the train, test and valid subsets as a DSet, which can be used to complete 
        the data preparation.
        
        A first call will trigger rows to be assigned to the train, test and valid part, which are stored
        in place to reproduce the exact same split for consecutive calls.

        Arguments:
            dataset: class (None)
                The DataSet class to use
        
        Returns: list(DataSet)
        """
        res = [ self.train.to_dataset(datasetclass) ]
        if len(self._test_indices) > 0:
            res.append(self.test.to_dataset(datasetclass))
        else:
            res.append(None)
        if len(self._valid_indices) > 0:
            res.append(self.valid.to_dataset(datasetclass))
        else:
            res.append(None)
        return res
        
    def df_to_dset(self, df):
        """
        Converts a DataFrame to a DSet that has the same pipeline as this DFrame.
        
        Arguments:
            df: DataFrame
        
        Returns: DSet
        """
        return self._dset(df)
        
    def to_databunch(self, datasetclass=None, batch_size=32, test_batch_size=None, 
                     num_workers=0, shuffle=True, pin_memory=False, collate=None, balance=None):
        """
        Prepare the data as a Databunch that contains dataloaders for the train, test and valid part.
        
        batch_size, num_workers, shuffle, pin_memory: see Databunch/Dataloader constructor.
        
        A first call will trigger rows to be assigned to the train, test, and valid part, which are stored
        in place to reproduce the exact same split for consecutive calls. 
        
        If the DFrame was only split in a train/test part, the test part is also used for validation. This
        is done because training Neural Networks always requires a validation set.

        Returns: Databunch
        """
        train, test, valid = self.to_datasets(datasetclass=datasetclass)
        if valid is None:
            valid = test
        return Databunch(self, train, test, valid, 
                         batch_size=batch_size, test_batch_size=test_batch_size, 
                         num_workers=num_workers, shuffle=shuffle, 
                         pin_memory=pin_memory, collate=collate, balance=balance)    

    def to_dataloaders(self, datasetclass=None, batch_size=32, test_batch_size=None, 
                     num_workers=0, shuffle=True, pin_memory=False, collate=None, balance=None):
        """
        Prepare the data as a Databunch that contains dataloaders for the train, test and valid part.
        
        batch_size, num_workers, shuffle, pin_memory: see Databunch/Dataloader constructor.
        
        A first call will trigger rows to be assigned to the train, test, and valid part, which are stored
        in place to reproduce the exact same split for consecutive calls. 
        
        If the DFrame was only split in a train/test part, the test part is also used for validation. This
        is done because training Neural Networks always requires a validation set.

        Returns: (DataLoader, DataLoader, DataLoader) 
            Two or three torch.utils.data.DataLoader objects for the train, test and 
            validation sets. These DataLoaders can be used by a PyTorch trainer to
            train, validate and evaluate the model.
        """
        db = self.to_databunch(datasetclass=datasetclass, batch_size=batch_size, 
                               test_batch_size=test_batch_size, num_workers=num_workers, 
                               shuffle=shuffle, pin_memory=pin_memory, 
                               collate=collate, balance=balance)
        return db.train_dl, db.test_dl, db.valid_dl
    
    def _evaluator(self, *metrics):
        """
        Creates a PipeTorch Evaluator, that can be used to visualize the data, the results
        of a model or cache learning/validation diagnostics. Since datasets are often
        reused in repeated experiments, every call will create a new Evaluator to prevent
        mixing results from different experiments. If you do want to compare results from
        multiple experiments, create a single evaluator and reuse that for the experiments
        that you wish to compare.
        
        Arguments:
            *metrics: callable
                One or more functions, that will take y_true, y_pred as parameter to
                compute an evaluation metric. Typically, functions from SKLearn.metrics 
                can be used.
        
        returns: Evaluator
        """
        return EvaluatorSK(self, *metrics)
   
    def evaluator(self, *metrics):
        try:
            if self._pt_evaluator.metrics == metrics:
                return self._pt_evaluator
        except: pass
        self._pt_evaluator = self._evaluator(*metrics)
        return self._pt_evaluator

    def from_numpy(self, x, y=None):
        """
        Create a dataset from Numpy arrays
        
        Args:
            x: Numpy Array
                containing either only the feature vectors X, or combining X and y
            y: Numpy Array (None)
                if present, these will contain the labels
                
        Returns: DSet
        """
        import torch
        
        if y is not None:
            assert x.shape[1] == len(self._columnx) and y.shape[1] == len(self._columny), 'the columns of x and y must match the configured number of columns in de DFrame'     
            if len(y.shape) == 1:
                y = y.reshape((len(y), len(self._columny)))
        else:
            if x.shape[1] == len(self._columnx) + len(self._columny):
                y = x[:,-len(self._columny):]
                x = x[:,:-len(self._columny)]
            elif x.shape[1] == len(self._columnx):
                y = np.zeros((len(x), len(self._columny)))
            else:
                raise ValueError('x must either have as many columns in x or the entire df')
        if len(x.shape) == 1:
            x = x.T
        if type(x) == torch.Tensor:
            x = x.numpy()
        elif isinstance(x, (pd.Series, pd.DataFrame)):
            x = x.to_numpy()
        if len(y.shape) == 1:
            y = y.T
        if type(y) == torch.Tensor:
            y = y.numpy()
        elif isinstance(y, (pd.Series, pd.DataFrame)):
            y = y.to_numpy()
        series = [ pd.Series(x[:,i], name=s) for i, s in enumerate(self._columnx)]
        series.extend([ pd.Series(y[:,i], name=s) for i, s in enumerate(self._columny) ] )
        df = pd.concat(series, axis=1)
        return self._dset(df)
       
    def from_list(self, x):
        return self.from_numpy(np.array(x))
    
    def _dset_indices(self, indices, transforms):
        if self._pt_sequence_window is None:
            return self._dset(self.loc[indices], transforms=transforms)
        else:
            try:
                low, high = min(indices), max(indices) + self._sequence_window + self._shift_y - 1
                return self._dset(self.loc[low:high], transforms=transforms)
            except:
                return self._dset(self.loc[:0], transforms=transforms)
    
    @property
    def train(self):
        """
        Prepares the train subset as a DSet and optionally trains transformation parameters, e.g. for image
        normalization or text tokenization. Transformations are only supported for PyTorch DataSets.
        
        A first call will trigger rows to be assigned to the train, valid and test part, which are stored
        in place to reproduce the exact same split for consecutive calls.
        
        Returns: DSet
        """
        if self._unchanged() and self._cached_train is not None:
            return self._cached_train
        
        r = self._dset_indices(self._train_indices, self._dataset_transforms())
        if self._dataset_train_transformation_parameters(r):
            r = self._dset_indices(self._train_indices, self._dataset_transforms())
        self._change('_cached_train', r)
        self._cached_raw_train = None
        return r
                    
    @property
    def raw_train(self):
        if self._unchanged() and self._cached_raw_train is not None:
            return self._cached_raw_train
        self._cached_raw_train = self._dset_indices(self._train_indices, self._dataset_transforms())
        return self._cached_raw_train
    
    @property
    def valid(self):
        if self._unchanged() and self._cached_valid is not None:
            return self._cached_valid
        
        r = self._dset_indices(self._valid_indices, self._dataset_transforms(train=False))
        self._change('_cached_valid', r)
        return r

    @property
    def test(self):
        if self._unchanged() and self._cached_test is not None:
            return self._cached_test
        
        if self._pt_sequence_window is None:
            r = DSet.df_to_testset(self.loc[self._test_indices], self, self._dataset_transforms(train=False))
        else:
            low, high = min(self._test_indices), max(self._test_indices) + self._sequence_window + self._shift_y - 1
            r = DSet.df_to_testset(self.loc[low:high], self, self._dataset_transforms(train=False))
        self._change('_cached_test', r)
        return r
    
    @property
    def train_X(self):
        return self.train.X
            
    @property
    def train_y(self):
        return self.train.y

    @property
    def valid_X(self):
        return self.valid.X
       
    @property
    def valid_y(self):
        return self.valid.y

    @property
    def test_X(self):
        return self.test.X
    
    def numpy(self):
        """
        Returns: train_X, train_y, valid_X, valid_y
        """
        return self.train_X, self.train_y, self.valid_X, self.valid_y
        
    def tensor(self):
        return self.train.X_tensor, self.train.y_tensor, self.valid.X_tensor, self.valid.y_tensor
    
    @property
    def test_y(self):
        return self.test.y
    
    def loss_surface(self, model, loss, **kwargs):
        self._evaluator(loss).loss_surface(model, loss, **kwargs)
    
    def inverse_scale_y(self, y):
        """
        Inversely scale an output y vector.
        
        Arguments:
        y: Numpy array or PyTorch tensor
            with the output that where preprocessed by this DFrame or predicted by the model
        
        Return: Pandas DataFrame
            That is reconstructed from Numpy arrays/Pytorch tensors
            that are transformed back to the original scale. 
        """
        y = to_numpy(y)
        if len(y.shape) == 1:
            y = y.reshape(-1,1)
        df = pd.DataFrame(self._inverse_scale(to_numpy(y), self._scalery, self._columny))
        return df
    
    def inverse_y(self, y):
        """
        Invert the transformation an output y vector, both scaling and categories.
        
        Arguments:
        y: Numpy array or PyTorch tensor
            with the output that where preprocessed by this DFrame or predicted by the model
        
        Return: Pandas DataFrame
            That is reconstructed from Numpy arrays/Pytorch tensors
            that are transformed back to the original scale and/or categorylabels. 
        """
        df = self.inverse_scale_y(y)
        if self._categoryy() is not None:
            for c, cat in zip(self._columny, self._categoryy()):
                if cat is not None:
                    df[c] = cat.inverse_transform(df[c])
        return df

    def add_column(self, y, indices, erase_y=True, columns=None):
        """
        Intended for internal use. Adds a column with values for the target variable to the DataFrame. 
        When applicable, the transformation for the target variable is automatically inverted. 
        This is useful to evaluate or visualize results.
        
        This effect is not inplace, but configured to a copy that is returned. 
        
        Arguments:
            y: Numpy/PyTorch array
                values in the same for as those generated for the target variable 
                (possibly predictions) that will be added to the DataFrame
            erase_y: bool (True)
                whether the original target variable is removed 
        
        Returns: DFrame 
        """
        df_y = self.inverse_scale_y(y)
        r = self.copy()
        if columns is None:
            columns = [ c + '_pred' for c in self._columny ]
        r[columns] = np.NaN
        r.loc[r.index[indices], columns] = df_y.values
        r = self._dframe(r)
        r._cached_changed = True
        return r
    
    def inverse_scale_X(self, X):
        """
        Inversely scale an input X matrix.
        
        Arguments:
        X: Numpy array or PyTorch tensor
            with the input features that where preprocessed by this DataFrame
        
        Return: Pandas DataFrame
            That is reconstructed from Numpy arrays/Pytorch tensors
            that are transformed back to the orignal scale. 
        """
        if self._pt_bias:
            X = X[:, 1:]
        if self._pt_polynomials is not None:
            X = X[:, :len(self._columnx)]
        df = self._inverse_scale(to_numpy(X), self._scalerx[:len(self._columnx)], self._columnx)
        if self._categoryx() is not None:
            for c, cat in zip(self._columnx, self._categoryx()):
                if cat is not None:
                    df[c] = cat.inverse_transform(df[c])
        return df

    def _inverse_scale(self, data, scalerlist, columns):
        data = to_numpy(data)
        if scalerlist is not None:
            data = [ data[:, i:i+1] if scaler is None else scaler.inverse_transform(data[:,i:i+1]) for i, scaler in enumerate(scalerlist) ]
        series = [ pd.Series(x.reshape(-1), name=c) for x, c in zip(data, columns)]
        return pd.concat(series, axis=1)

    def inverse_scale(self, X, y, y_pred = None, cum=None):
        """
        Reconstructs a DSet from Numpy arrays/Pytorch tensors
        that are scaled back to the original scale. 
        This is useful for evaluation or to visualize results.
        
        Arguments:
            X: Numpy array or PyTorch tensor 
                with the same format as input features that were generated by the DataFrame
                
            y: Numpy array or PyTorch tensor
                with the same format as the target variable that was generated by the
                DataFrame
                
            pred_y (optional): Numpy array or PyTorch tensor
                with the same format as the target variable that was generated by the
                DataFrame
                
            cum (optional): DSet
                an dataset to add the results to, 
                to accumulate predictions over several mini-batches.
        
        Returns: DSet 
        """
        y = self.inverse_scale_y(y)
        X = self.inverse_scale_X(X)
        if y_pred is not None:
            y_pred = self.inverse_scale_y(y_pred).add_suffix('_pred')
            df = pd.concat([X, y, y_pred], axis=1)
        else:
            df = pd.concat([X, y], axis=1)
        if cum is not None:
            df = pd.concat([cum, df])
        df = self._dset(df)
        return df
        
    def plot_boundary(self, predict):
        self._evaluator().plot_boundary(predict)

    def balance(self, weights=True):
        """
        Oversamples rows in the training set, so that the values of the target variable 
        are better balanced. Does not affect the valid/test set.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            weights: True or dict
                when set to True, the target values of the training set are 
                uniformely distributed,
                otherwise a dictionary can be passed that map target values to the 
                desired fraction of the training set (e.g. {0:0.4, 1:0.6}).
        
        Returns: DFrame
        """
        r = self.copy(deep=False)
        r._change('_pt_balance', weights)
        return r    
  
    def scale(self, columns=True, scalertype=StandardScaler, omit_interval=(-2,2)):
        """
        Scales the features and target variable in the DataFrame. A scaler is fitted on the
        train data and applied to train, valid and test.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            columns: True, str or list of str (True)
                the columns to scale (True for all)
                
            scalertype: a SKLearn type scaler (StandardScaler)
                the scaler class that is used
            
            omit_interval: (-2,2) when colums is set to True
                features whose values lie within this interval are not scaled, the default value
                ensures that binary values will not be scaled.
                
        Return: DFrame
        """
        if self._pt_polynomials and columns != 'x_only':
            assert type(columns) != list or len(columns) == 0, 'You cannot combine polynomials with column specific scaling'
        r = self.copy(deep=False)
        r._change('_pt_scale_columns', columns)
        r._change('_pt_scalertype', scalertype)
        r._change('_pt_scale_omit_interval', omit_interval)
        return r
    
    def scalex(self, scalertype=StandardScaler, omit_interval=(-2,2)):
        """
        Scale all input features. A scaler is fitted on the
        train data and applied to train, valid and test.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            scalertype: a SKLearn scaler class (StandardScaler)
                the scaler class that is used
                
            omit_interval: tupple (-2,2) 
                features whose values lie within this interval are not scaled, the default value
                ensures that binary values will not be scaled.
        
        Returns: DFrame 
        """
        return self.scale(columns='x_only', scalertype=scalertype, omit_interval=omit_interval)
    
    def add_bias(self):
        """
        Adds a bias column of value 1 to generated inputs.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Returns: DFrame
        """
        r = self.copy(deep=False)
        r._change('_pt_bias', True)
        return r
    
    def reshuffle(self, random_state=None):
        """
        Resamples the train, valid and test set with the existing settings.
        
        This effect is not inplace, but applied to a copy that is returned. 

        Arguments:
            random_state: int (None)
                set a random_state for reproducible results   
                
        Returns: DFrame
        """
        r = self.copy(deep=False)
        r._change('_pt_split_random_state', random_state)
        return r
    
    def _index_changed(self):
        for p in self._cached_indices:
            setattr(self, p, None)
        for p in self._cached_fold:
            setattr(self, p, None)
        for p in self._cached:
            setattr(self, p, None)
        self._cached_index = self.index.copy()
        
    def _columns_changed(self):
        for p in self._cached:
            setattr(self, p, None)
    
    def reset_indices(self):
        """
        Clears the currently sampled split() and folds(), which is stored whenever data preparation
        is called. Therefore, resampling the split or fold on the next call to data preparation.
        However, this call will not reset a random_state, use reshuffle() for that.
        
        This effect is not inplace, but applied to a copy that is returned. 
        
        Returns: DFrame
        """
        r = self.copy(deep=False)
        r._index_changed()
        return r
        
    def split(self, 
              test_size=None, 
              valid_size=None, 
              shuffle=None, 
              random_state=None, 
              stratify=None, 
              stratify_test=None):
        """
        Split the data in a train/valid/(test) set. Any existing split is reset.
        
        This effect is not inplace, but applied to a copy that is returned. 
        
        Arguments:
            test_size: float (None)
                the fraction of the dataset that is used for the test set. When combined with folds
                if 1 > test_size > 0, the test set is split before the remainder is divided in folds 
                to apply n-fold cross validation.
                
            valid_size: float (None)
                the fraction of the dataset that is used for the validation set.
                
            shuffle: bool (None)
                shuffle the rows before splitting. None means True unless sequence() is called to process
                the data as a (time) series.
                
            random_state: int (None)
                set a random_state for reproducible results
                
            stratify: str, [ str ], True or None (None)
                Apply stratified sampling over these columns (True = all columns)
                For multiple columns, multi-label stratification is applied with support over
                continuous variables.
                
            stratify_test: str, [ str ], True or (None)
                Apply stratified sampling over these columns (True = all columns)
                For multiple columns, multi-label stratification is applied with support over
                continuous variables.
                If None, the value for stratify is used. To supress stratification, pass [].
            
        Returns: DFrame 
        """
        r = self.copy(deep=False)
        assert test_size is None or self._pt_test_indices is None, 'A fixed test set was given, therefore you cannot use test_size'
        
        r._change('_pt_valid_size', valid_size)
        r._change('_pt_test_size', test_size)
        r._change('_pt_split_shuffle', shuffle)
        r._change('_pt_split_random_state', random_state)
        r._change('_pt_split_stratify', [stratify] if type(stratify) == str else stratify)
        r._change('_pt_split_stratify_test', [stratify_test] if type(stratify_test) == str else \
                                    (r._pt_split_stratify if stratify_test is None else stratify_test))
        r._index_changed()
        return r

    def fixtest(self):
        """
        Returns a copy of this DFrame in which the test set is split and fixed, so that
        any consecutive splits and folds are only applied to the train/valid set.
        """
        t = self.copy(deep=False)
        r = t.from_train_test(self.train, self.test)
        t._copy_non_cached_meta(r)
        r._pt_test_size = None
        r._pt_split_stratify_test = None
        return r
    
    def folds(self, folds=5, shuffle=True, random_state=None, stratify=None, mode='both'):
        """
        Divide the data in folds to setup n-Fold Cross Validation in a reproducible manner. 
        
        By combining folds() with split(0 < test_size < 1) , a single testset is split before 
        dividing the remainder in folds that are used for training and validation. 
        When used without split, by default a single fold is used for testing.
        
        The folds assigned to the validation and test-set rotate differently, 
        giving 5x4 combinations for 5-fold cross validation. You can access all 20 combinations
        by calling fold(0) through fold(19).
    
        This effect is not inplace, but applied to a copy that is returned. 
    
        Arguments:
            folds: int (None)
                The number of times the data will be split in preparation for n-fold cross validation. The
                different splits can be used through the fold(n) method.
                SKLearn's SplitShuffle is used, therefore no guarantee is given that the splits are
                different nor that the validation splits are disjoint. For large datasets, that should not
                be a problem.
                
            shuffle: bool (None)
                shuffle the rows before splitting. None means True unless sequence() is called to process
                the data as a (time) series.
                
            random_state: int (None)
                set a random_state for reproducible results.
                
            stratify: str, [ str ], True or None (None)
                Apply stratified sampling over the given columns. True means all columns. If column 
                Per value for the given column, the rows are sampled. When a list
                of columns is given, multi-label stratification is applied.
                
            mode: ['both', 'test'] (both)
                when set to 'both', on every split one fold is used for validation, 
                one fold for testing and the remainder for training.
                when set to 'both' with a fixed test set, one fold is used for validation,
                and the remainder for traing ( and the fixed test set is kept separate )
                when set to 'test' one fold is used for testing.
            
        Returns: copy of DFrame 
            schedules the data to be split in folds.
        """
        assert type(folds) == int and folds > 1, 'You have to set split(folds) to an integer > 1'
        assert mode == 'both' or mode == 'test', 'mode has to be both or test'
        assert mode == 'both' or len(self.fixed_test_indices) == 0, 'It does not make sense to use mode=test with a fixed testset'
        r = self.copy(deep=False)
        r._change('_pt_folds', folds)
        r._change('_pt_folds_shuffle', shuffle)
        r._change('_pt_folds_random_state', random_state)
        r._change('_pt_folds_stratify', [stratify] if type(stratify) == str else stratify)
        r._pt_fold = 0
        if mode == 'both':
            if len(self.fixed_test_indices) > 0:
                r._change('_pt_folds_mode', 'valid')
            else:
                r._change('_pt_folds_mode', 'both')
        else:
            r._change('_pt_folds_mode', 'test')
        return r
    
    def leave_one_out(self):
        """
        Configures folds() to perform a leave-one-out cross validation.
        """
        return self.folds(len(self._indices_after_testsplit), shuffle=False)
        
    def polynomials(self, degree, include_bias=False):
        """
        Adds (higher-order) polynomials to the data pipeline.
        
        This effect is not inplace, but configured to a copy that is returned. 
        
        Arguments:
            degree: int - degree of the higher order polynomials (e.g. 2 for squared)
            include_bias: bool (False) - whether to generate a bias column
        
        Returns: copy of DFrame 
        """
        assert type(self._pt_scale_columns) != list or len(self._pt_scale_columns) == 0, 'You cannot combine polynomials with column specific scaling'
        r = self.copy(deep=False)
        r._change('_pt_polynomials', PolynomialFeatures(degree, include_bias=include_bias))
        return r
    
    def no_columny(self):
        """
        PipeTorch cannot currently handle a dataset without a target variable, 
        however, it can work by simply assigning one of the used input features
        as a target variable.

        This effect is not inplace, but configured to a copy that is returned. 

        Returns: copy of DFrame
        """
        r = self.copy(deep=False)
        r._change('_pt_columny', [self._columnx[0]])
        return r
    
    def columny(self, columns=None, vector=None):
        """
        Configures the generated target variable and shape.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            columns: str or list of str (None)
                single column name or list of columns that is to be used as target column. 
                None: use the last column
            vector: bool (None)
                By default, y will be prepared as a 1d vector and y_tensor as an (n, 1) matrix
                because most SKLearn algorithmms prefer vector and for PyTorch (n, 1) matrices 
                are used more often. Setting vector=False causes both y and y_tensor to return 1d shapes
                and vector=True causes both y and y_tensor to return (n, 1) shapes.
        
        Returns: DFrame 
        """
        
        r = self.copy(deep=False)
        if columns is not None:
            r._change('_pt_columny', [columns] if type(columns) == str else columns)
        assert r._pt_columny is None or len(r._pt_columny) == 1 or not vector, 'You cannot create target vector with multiple columns'
        r._change('_pt_vectory', vector)
        return r

    def columnx(self, *columns, omit=False):
        """
        Specify which columns to use as input. The target variable is always excluded
        so if you want to add that (for sequence learning), you should copy the
        target column. 
        Unlocks the DFrame, train and valid sets are therefore resampled.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            columns: str or list of str
                columns to be used as input features
            omit: bool (False)
                when True, all columns are used except the specified target column(s)
        
        Returns: DFrame
        """
        r = self.copy(deep=False)
        if omit:
            r._change('_pt_columnx', [ c for c in self.columns if c not in columns ])
        else:
            r._change('_pt_columnx', list(columns) if len(columns) > 0 else None)
        return r
    
    def category(self, *columns, sort=False):
        """
        Converts the values in the targetted columns into indices, for example to use in lookup tables.
        columns that are categorized are excluded from scaling. You cannot use this function together
        with polynomials or bias.
        
        Note: PipeTorch only uses categories that are in the training set and uses category 0 as
        an unknown category number for categories in the validation and test set that are not known during training.
        This way, no future information is used. 
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            columns: str or list of str
                list of columns that is to be converted into a category
            sort: True/False (default False) 
                whether the unique values of these colums should be converted 
                to indices in sorted order.
        
        Returns: DFrame
        """
        assert self._pt_polynomials is None, 'You cannot combine categories with polynomials'
        assert self._pt_bias is None, 'You cannot combine categories with polynomials'
        r = self.copy(deep=False)
        r._change('_pt_category', columns)
        r._change('_pt_category_sort', sort)
        return r
    
    def dummies(self, *columns):
        """
        Converts the values in the targetted columns into dummy variables. This is an alernative to 
        pd.get_dummies, that only uses the train set to assess which values there are (as it should be),
        setting all variables to 0 for valid/test items that contain an unknown label. 
        That way, no future information is used. Columns that are categorized are excluded from scaling. 
        You cannot use this function together with polynomials or bias.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Args:
            columns: str or list of str
                the columns that are to be converted into a category
        
        Returns: DFrame 
        """
        assert self._pt_polynomials is None, 'You cannot combine categories with polynomials'
        assert self._pt_bias is None, 'You cannot combine categories with polynomials'
        r = self.copy(deep=False)
        r._change('_pt_dummies', columns)
        return r

    def sequence(self, window, shift_y = 1):
        """
        The rows in the DataFrame are considered to be a continuous sequence over time. 
        From this sequence, input samples are created that contain the features over a 'window'
        of rows. Ths allows to learn a model to predict a target based on prior history. The
        samples are generated so that their window overlaps.
        Unlocks the DFrame, train and valid sets are therefore resampled.
        
        Samples with NaN's are automatically skipped. When DataFrames are grouped, samples will
        be created only within groups.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            window: int
                the number of rows that is used a an input
            shift_y: int
                how many rows into the future the target variable is placed, 
                e.g. with a window=2 and shift_y=1, X0 would contain [x[0], x[1]] 
                while y0 would contain y[2], 
                the next sample X1 would be [x[1], x[2]] and y[3].
        
        Returns: DFrame
        """
        r = self.copy(deep=False)
        r._change('_pt_sequence_window', window)
        r._change('_pt_sequence_shift_y', shift_y)
        return r
        
    def filterna(self, filter=True):
        """
        Returns: a DataFrame in which rows with missing values are filtered
        from the train and valid set. The result is not visible in the DataFrame
        but will be in the resulting train and valid sets.
        
        Args:
            filter: bool (True)
                Default is to drop rows in de DSet's that still have missing values.
                Set to false to turn off
        """
        r = self.copy(deep=False)
        r._change('_pt_filterna', filter)
        return r
    
    def columnfunction(self, **columnfunctions):
        """
        Returns: a DataFrame in which column transformations are added to the
        data pipeline, before scaling the data.
        
        Transforms effectively adds to the existing configured column functions,
        while overriding functions that are newly defined. This function does not
        remove any column functions, unless called with zero parameters
        
        Args:
            columnfunctions: { columname: callable}
                when empty, all existing column functions will be erased,
                otherwise, the given column functions are added to the dictionary
                of configured column functions, which are executed in the pipeline
                just before str column names and a function
                that will be performed on the indicated column.
        """

        assert all([ c in self.columns for c in columnfunctions ]), 'Not all keys are columns'
        r = self.copy(deep=False)
        if r._pt_transform is None:
            r._change('_pt_transform', columnfunctions)
        else:
            r._change('_pt_transform', r._pt_transform | columnfunctions)
        return r
    
    def log(self, *columns):
        """
        Returns: a DataFrame in which log column transformations are added to the
        data pipeline, before scaling the data.
        
        Args:
            columnfunctions: { columname: callable}
                a dictionary of str column names and a function
                that will be performed on the indicated column.
        """
        return self.columnfunction( **{c:FunctionTransformer(np.log) for c in columns })
        
    def log1p(self, *columns):
        """
        Returns: a DataFrame in which log1p column transformations are added to the
        data pipeline, before scaling the data.
        
        Args:
            columnfunctions: { columname: callable}
                a dictionary of str column names and a function
                that will be performed on the indicated column.
        """
        return self.columnfunction( **{c:FunctionTransformer(np.log1p) for c in columns })
    
    def dataset_train_transforms(self, *transforms):
        """
        Configure a (list of) transformation function(s) that is called from the DataSet class to prepare the 
        train data.
        
        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            *transforms: [ callable ]
                (list of) transformation function(s) that is called from the DataSet class to prepare the data.

        Returns: DFrame
        """
        r = self.copy(deep=False)
        r._change('_pt_dataset_train_transforms', transforms)
        return r
    
    def dataset_transforms(self, *transforms):
        """
        Configure a (list of) transformation function(s) that is called when retrieving
        items from PipeTorch' TransformableDataSet class to prepare the data. These function allow
        to configure a pipeline for data augmentation that is used to train PyTorch models.
        For example, to train of images or audio fragments that are read from disk.
        
        Note that these transformations are not used to prepare Numpy Arrays, and the DataSet class has
        to call the tranformations (which the TransformableDataSet class does that is used by default).

        This effect is not inplace, but configured to a copy that is returned. 

        Arguments:
            *transforms: [ callable ]
                (list of) transformation function(s) that is called from the DataSet class to prepare the data.
                
        Returns: DFrame
        """
        r = self.copy(deep=False)
        r._change('_pt_dataset_transforms', transforms)
        return r

    def _dataset_train_transformation_parameters(self, train_dset):
        """
        Placeholder to extend DFrame to configure transformations that are applied on the train DataSet,
        but not on the validation or test set. This is used by
        ImageDFrame to learn normalization parameters for images and to learn a vocabulary 
        for TextDFrame to tokenize text.
        
        Note that the mechanism assumes that the train DSet is generated first.
        
        Arguments: 
            train_dset: DSet
                The train DSet to learn the transformation parameters on
            
        Returns: bool (None)
            This function should return True when parameters were learned, to force DFrame to add a new
            list of transformations to the generated DSet.
        """
        pass
    
    def _dataset_pre_transforms(self):
        """
        Placeholder for standard transformations that precede configured transformations.
        """
        return []
    
    def _dataset_post_transforms(self):
        """
        Placeholder for standard transformations that succede configured transformations.
        """
        return []
    
    def _dataset_transforms(self, pre=True, train=True, standard=True, post=True):
        """
        Arguments:
            pre: bool (True) - whether to include pre transformations
            train: bool (True) - whether to include train transformations
            standard: bool (True) - whether to include standard transformations
            post: bool (True) - whether to include post transformations
            
        Returns: [ callable ]
            A list of transformations that is applied to the generated DSet 
        """
        t = []
        try:
            if pre:
                t.extend(self._dataset_pre_transforms())
        except: pass
        try:
            if train:
                t.extend(self._pt_dataset_train_transforms)
        except: pass
        try:
            t.extend(self._pt_dataset_transforms)
        except: pass
        try:
            if post:
                t.extend( self._dataset_post_transforms() )
        except: pass
        return t
    
    def inspect(self):
        """
        Describe the data in the DFrame, specifically by reporting per column 
        - Datatype 
        - Missing: number and percetage of 'Missing' values
        - Range: numeric types, are described as a range [min, max]
                  and for non-numeric types the #number of unique values is given
        - Values: the two most frequently occuring values (most frequent first).
        """

        missing_count = self.isnull().sum() # the count of missing values
        value_count = self.isnull().count() # the count of all values
        missing_percentage = round(missing_count / value_count * 100,2) #the percentage of missing values

        datatypes = self.dtypes

        df = pd.DataFrame({
            'Missing (#)': missing_count, 
            'Missing (%)': missing_percentage,
            'Datatype': datatypes
        }) #create a dataframe
        df = df.sort_values(by=['Missing (#)'], ascending=False)

        value_col = []
        range_col = []
        for index, row in df.iterrows():
            u = self[index].value_counts().index.tolist()
            if pd.api.types.is_numeric_dtype(row['Datatype']):
                _range = f"[{self[index].min()}, {self[index].max()}]"
            else:
                _range = f"#{len(u)}"
            if len(u) == 1:
                _values = f'({u})'
            elif len(u) == 2:
                _values = f'({u[0]}, {u[1]})'
            elif len(u) > 2:
                _values = f'({u[0]}, {u[1]}, ...)'
            else:
                _values = ''
            range_col.append(_range)
            value_col.append(_values)
        df["Range"] = range_col
        df["Values"] = value_col
        return df

    def study(self, *target, evaluator=None, **kwargs):
        """
        On a DFrame that is configured for n-fold cross validation (df.folds(n)), this function iterates
        over the trials, fitting an SKLearn model and storing the targets for the train, valid and test subsets.
        
        Arguments:
            target: callable
                one or more functions that return an evaluation metrics when called with
                target(y_true, y_predict)
            evaluator: Evaluator (None)
                when provided, the results are added to this evaluator, otherwise
                the evaluator of this DFrame is reset and a new one is used.
        """
        if evaluator is None:
            try:
                del self._pt_evaluator
            except: pass
            evaluator = self.evaluator(*target)
        study = evaluator.study(**kwargs)
        return study

class DFrame(pd.DataFrame, _DFrame):
    _metadata = _DFrame._metadata
    _internal_names = _DFrame._internal_names
    _internal_names_set = _DFrame._internal_names_set

    def __init__(self, data, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        _DFrame.__init__(self, data)
        self._catch_change()

    def _changed(self, f, *args, **kwargs):
        #print(f'changed {f} {args} {kwargs}')
        self._columns_changed()
        return f(*args, **kwargs)

    def _return_changed(self, f, *args, **kwargs):
        r = f(*args, **kwargs)
        r._columns_changed()
        return r

    def _change(self, key, value):
        #print(f'changed_attr {key} {value}')
        if key in self._config_indices_set:
            self._index_changed()
        elif key in self._config_set:
            self._columns_changed()
        return super().__setattr__(key, value)

    def __changed_inplace(self, **kwargs):
        try:
            if kwargs['inplace']:
                if r._unchanged():
                    r._columns_changed()
        except: pass
        try:
            if kwargs['copy'] == False:
                if r._unchanged():
                    r._columns_changed()
        except: pass

    
    def _changed_inplace(self, f, *args, **kwargs):
        #print(f'changed_inplace {f} {args} {kwargs}')
        self.__changed_inplace(**kwargs)
        r = f(*args, **kwargs)
        r._columns_changed()
        return r
     
    def _unchanged(self):
        #print(f'process_changed {self._cached_changed} {self.index.equals(self._cached_index)}')
        
        if self._cached_index is not None and self.index.equals(self._cached_index):
            return True
        self._index_changed()
        return False
    
    def _catch_change_f(self, f):
        return functools.wraps(f)(functools.partial(self._changed, f))
    
    def _catch_return_change_f(self, f):
        return functools.wraps(f)(functools.partial(self._return_changed, f))
    
    def _catch_change_inplace_f(self, f):
        return functools.wraps(f)(functools.partial(self._changed_inplace, f))
    
    def _catch_change(self):
        self._set_value = self._catch_change_f(self._set_value)
        self._setitem_slice = self._catch_change_f(self._setitem_slice)
        self._setitem_frame = self._catch_change_f(self._setitem_frame)
        self._setitem_array = self._catch_change_f(self._setitem_array)
        self._set_item = self._catch_change_f(self._set_item)
        self.__getitem__ = self._catch_return_change_f(self.__getitem__)
        self.update = self._catch_change_f(self.update)
        self.insert = self._catch_change_f(self.insert)

        self._mgr.idelete = self._catch_change_f(self._mgr.idelete)
        self._mgr.iset = self._catch_change_f(self._mgr.iset)
        self._mgr.insert = self._catch_change_f(self._mgr.insert)
        #self._mgr.column_setitem = self._catch_change_f(self._mgr.column_setitem)
        #self._mgr.reindex_axis = self._catch_change_f(self._mgr.reindex_axis)
        #self._mgr.reindex_indexer = self._catch_change_f(self._mgr.reindex_indexer)

        self.clip = self._catch_change_inplace_f(self.clip)
        self.drop = self._catch_change_inplace_f(self.drop)
        self.drop_duplicates = self._catch_change_inplace_f(self.drop_duplicates)
        #self.dropna = self._catch_change_inplace_f(self.dropna)
        self.eval = self._catch_change_inplace_f(self.eval)

        self.fillna = self._catch_change_inplace_f(self.fillna)
        self.interpolate = self._catch_change_inplace_f(self.interpolate)
        self.mask = self._catch_change_inplace_f(self.mask)
        self.pad = self._catch_change_inplace_f(self.pad)
        self.query = self._catch_change_inplace_f(self.query)
        self.replace = self._catch_change_inplace_f(self.replace)
        self.reset_index = self._catch_change_inplace_f(self.reset_index)
        self.set_axis = self._catch_change_inplace_f(self.set_axis)
        self.set_index = self._catch_change_inplace_f(self.set_index)
        self.sort_index = self._catch_change_inplace_f(self.sort_index)
        self.sort_values = self._catch_change_inplace_f(self.sort_values)
        self.where = self._catch_change_inplace_f(self.where)
        self.astype = self._catch_change_inplace_f(self.astype)
        self.reindex = self._catch_change_inplace_f(self.reindex)
        self.rename = self._catch_change_inplace_f(self.rename)
        #self.__setattr__ = self._changed_attr

    def dropna(self, **kwargs):
        if self._pt_test_indices is not None:
            warnings.warn("Using dropna on a dataset with a fixed test set " +
                    "which is probably a bad idea since this removes test data.",
                    RuntimeWarning,
                )
        self.__changed_inplace(**kwargs)
        return pd.DataFrame.dropna(self, **kwargs)

    @classmethod
    def read_csv(cls, path, **kwargs):
        df = pd.read_csv(path, **kwargs)
        return cls(df)

    @classmethod
    def from_dfs(cls, *dfs, **kwargs):
        return cls(pd.concat(dfs), **kwargs)
    
    @classmethod
    def from_train_test(cls, train, test, **kwargs):
        r = cls(pd.concat([train, test], ignore_index=True))
        r._pt_train_test_indices = list(range(len(train)))
        r._pt_test_indices = list(range(len(train), len(train)+len(test)))
        return r
    
    @classmethod
    def read_from_kaggle(cls, dataset, train=None, test=None, shared=True, force=False, **kwargs):
        """
        Reads a DFrame from a Kaggle dataset. The downloaded dataset is automatically stored so that the next time
        it is read from file rather than downloaded. See `read_csv`. The dataset is stored by default in a folder
        with the dataset name in `~/.pipetorchuser`.

        If the dataset is not cached, this functions requires a valid .kaggle/kaggle.json file, that you can 
        create manually or with the function `create_kaggle_authentication()`.

        Note: there is a difference between a Kaggle dataset and a Kaggle competition. For the latter, 
        you have to use `read_from_kaggle_competition`.

        Example:
            read_from_kaggle('uciml/autompg-dataset')
                to read/download `https://www.kaggle.com/datasets/uciml/autompg-dataset`
            read_from_kaggle('robmarkcole/occupancy-detection-data-set-uci', 'datatraining.txt', 'datatest.txt')
            to combine a train and test set in a single DFrame

        Arguments:
            dataset: str
                the username/dataset part of the kaggle url, e.g. uciml/autompg-dataset for 

            train: str (None)
                the filename that is used as the train set, e.g. 'train.csv'
            test: str (None)
                the filename that is used as the test set, e.g. 'test.csv'
            shared: bool (False)
                save the dataset in ~/.pipetorch instead of ~/.pipetorchuser, allowing to share downloaded
                files between users.
            force: bool (False)
                when True, the dataset is always downloaded
            **kwargs:
                additional parameters passed to pd.read_csv. For example, when a multichar delimiter is used
                you will have to set engine='python'.

        Returns: DFrame
        """
        k = Kaggle(dataset, shared=shared)
        if force:
            k.remove_user()
        train = k.read(train, **kwargs)
        if test is not None:
            test = k.read(test, **kwargs)
            return cls.from_train_test(train, test)
        return cls(train)
           
#     @classmethod
#     def read_from_kaggle_competition(cls, dataset, train=None, test=None, shared=False, force=False, **kwargs):
#         train = read_from_kaggle_competition(dataset, filename=train, shared=shared, force=force, **kwargs)
#         if test is not None:
#             test = read_from_kaggle_competition(dataset, filename=test, **kwargs)
#             return cls.from_train_test(train, test)
#         return cls(train)

#     @classmethod
#     def read_csv(cls, url, filename=None, path=None, save=False, **kwargs):
#         """
#         Reads a .csv file from cache or url. The place to store the file is indicated by path / filename
#         and when a delimiter is used, this is also used to save the file so that the original delimiter is kept.
#         The file is only downloaded using the url if it does not exsists on the filing system. If the file is
#         downloaded and save=True, it is also stored for future use.

#         Arguments:
#             url: str
#                 the url to download or a full path pointing to a .csv file
#             filename: str (None)
#                 the filename to store the downloaded file under. If None, the filename is extracted from the url.
#             path: str (None)
#                 the path in which the file is stored. If None, it will first check the ~/.pipetorch (for sharing
#                 dataset between users) and then ~/.pipetorchuser (for user specific caching of datasets).
#             save: bool (False)
#                 whether to save a downloaded .csv
#             **kwargs:
#                 additional parameters passed to pd.read_csv. For example, when a multichar delimiter is used
#                 you will have to set engine='python'.

#         Returns: DFrame
#         """
#         return cls(read_csv(url, filename=filename, path=path, save=save, **kwargs))
   
    @classmethod
    def read_from_package(cls, package, filename, **kwargs):
        return cls(read_from_package(package, filename, **kwargs))
    
    @classmethod
    def read_from_function(cls, filename, function, path=None, save=True, **kwargs):
        """
        First checks if a .csv file is already stored, otherwise, calls the custom function to retrieve a 
        DataFrame. 

        The place to store the file is indicated by path / filename.
        The file is only retrieved from the function if it does not exsists on the filing system. 
        If the file is retrieved and save=True, it is also stored for future use.

        Arguments:
            filename: str (None)
                the filename to store the downloaded file under.
            function: func
                a function that is called to retrieve the DataFrame if the file does not exist.
            path: str (None)
                the path in which the file is stored. If None, it will first check the ~/.pipetorch (for sharing
                dataset between users) and then ~/.pipetorchuser (for user specific caching of datasets).
            save: bool (True)
                whether to save a downloaded .csv
            **kwargs:
                additional parameters passed to pd.read_csv. For example, when a multichar delimiter is used
                you will have to set engine='python'.

        Returns: DFrame
        """
        return cls(read_from_function(filename, function, path=path, save=save, **kwargs))

    @classmethod
    def read_excel(cls, path, filename=None, **kwargs):
        return cls(read_excel(path, filename=filename, **kwargs))

    @property
    def _constructor(self):
        return DFrame
    
    def groupby(self, by, axis=0, level=None, as_index=True, sort=True, group_keys=True, observed=False, dropna=True):
        """
        Groups a DFrame just like a Pandas DataFrame. This is useful for 
        learning sequences, since samples are only created using a sliding window
        within groups.
        
        Arguments:
            by: single column or list of columns
            for the other arguments, see pandas.DataFrame.groupby 
        
        Returns: pipetorch.PTGroupedDFrame
        """
        r = super().groupby(by, axis=axis, level=level, as_index=as_index, sort=sort, group_keys=group_keys, observed=observed, dropna=dropna)
        return self._copy_meta( GroupedDFrame(r) )

     
class DSeries(pd.Series, _DFrame):
    _metadata = _DFrame._metadata
    _internal_names = _DFrame._internal_names
    _internal_names_set = _DFrame._internal_names_set

    def __init__(self, data, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        _DFrame.__init__(self, data)

    @property
    def _constructor(self):
        return DSeries
    
    @property
    def _constructor_expanddim(self):
        return DFrame
    
class GroupedDSeries(SeriesGroupBy, _DFrame):
    _metadata = _DFrame._metadata
    _internal_names = _DFrame._internal_names
    _internal_names_set = _DFrame._internal_names_set

    def __init__(self, data, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        _DFrame.__init__(self, data)

    @property
    def _constructor(self):
        return GroupedDSeries
    
    @property
    def _constructor_expanddim(self):
        return GroupedDFrame
    
class GroupedDFrame(DataFrameGroupBy, _DFrame):
    _metadata = _DFrame._metadata
    _internal_names = _DFrame._internal_names
    _internal_names_set = _DFrame._internal_names_set

    def __init__(self, data=None):
        super().__init__(obj=data.obj, keys=data.keys, axis=data.axis, level=data.level, grouper=data.grouper, exclusions=data.exclusions,
                selection=data._selection, as_index=data.as_index, sort=data.sort, group_keys=data.group_keys,
                observed=data.observed, dropna=data.dropna)
        _DFrame.__init__(self, data)

    @property
    def _constructor(self):
        return GroupedDFrame
    
    @property
    def _constructor_sliced(self):
        return GroupedDSeries
    
    def _unchanged(self):
        return True
    
    def _index_changed(self):
        pass
        
    def _columns_changed(self):
        pass
    
    def get_group(self, namel, obj=None):
        return self._dframe( super().get_group(name, obj=obj) )
        
    def __iter__(self):
        for group, subset in super().__iter__():
            yield group, self._copy_meta(subset)
        
    def to_dataset(self):
        """
        Convert a grouped DFrame into a PyTorch ConcatDataset over datasets
        for every group contained.
        
        returns: train, valid, test DataSet
        """
        from torch.utils.data import ConcatDataset
        dss = []
        for key, group in self:
            dss.append( self.df_to_dataset(group) )

        return [ConcatDataset(ds) for ds in zip(*dss)]



from .trainer import Trainer, to_numpy, Trainer as trainer
import torch
from torch import nn
from torch.optim import *
from torch.optim.lr_scheduler import *
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from torcheval.metrics import *
from sklearn.metrics import *



from .textcollection import TextCollection
from .dframe import DFrame
from .helper import path_shared, path_user

def read_torchtext(torchtext_function):
    try:
        return torchtext_function(root=path_shared() )
    except:
        return torchtext_function(root=path_user() )

def wine_quality():
    return DFrame.read_from_kaggle('uciml/red-wine-quality-cortez-et-al-2009')

def telco_churn():
    return DFrame.read_from_kaggle('blastchar/telco-customer-churn')

def movielens_ratings():
    COLS = ['user_id', 'movie_id', 'rating', 'timestamp']
    return DFrame.read_from_kaggle('odedgolden/movielens-1m-dataset', 'ratings.dat', sep='::', engine='python', encoding = "ISO-8859-1", names=COLS)

def movielens_movies():
    COLS = ['movie_id', 'title', 'genre']
    return DFrame.read_from_kaggle('odedgolden/movielens-1m-dataset', 'movies.dat', sep='::', engine='python', encoding = "ISO-8859-1", names=COLS)
    
def movielens_users():
    COLS = ['user_id', 'gender', 'age', 'occupation', 'zip_code']
    return DFrame.read_from_kaggle('odedgolden/movielens-1m-dataset', 'users.dat', sep='::', engine='python', encoding = "ISO-8859-1", names=COLS)
    
def dam_outflow(**kwargs):
    return DFrame.read_from_package('pipetorch', 'data/datasets/dam_outflow.csv', **kwargs)
    
def realestate(**kwargs):
    return DFrame.read_from_package('pipetorch', 'data/datasets/realestate.csv', **kwargs)
    
def heart_disease():
    return DFrame.read_from_kaggle('mlms1', 'train.csv', 'test_set.csv')
    
def boston_housing_prices():
    return DFrame.read_from_kaggle('fedesoriano/the-boston-houseprice-data')

def hotel_booking():
    return DFrame.read_from_kaggle('mojtaba142/hotel-booking')

def hotel():
    df = hotel_booking()
    return df[(df.ArrivalDateYear < 2017) | (df.ArrivalDateWeekNumber < 14)]

def hotel_test_orig():
    df = hotel_booking()
    return df[(df.ArrivalDateYear == 2017) & (df.ArrivalDateWeekNumber > 13)]

def hotel_test():
    return hotel_test_orig().drop(columns='IsCanceled')

def hotel_test_y():
    return hotel_test_orig()[['IsCanceled']]

def hotel_test_score(pred_y):
    return f1_score(hotel_test_y(), pred_y)

def iris():
    return DFrame.read_from_kaggle('uciml/iris').drop(columns='Id')

def bank_marketing():
    return DFrame.read_from_kaggle('janiobachmann/bank-marketing-dataset')

def auto_mpg():
    return DFrame.read_from_kaggle('uciml/autompg-dataset')

def big_mart_sales():
    return DFrame.read_from_kaggle('brijbhushannanda1979/bigmart-sales-data', train='Train.csv', test='Test.csv')

def advertising_sales():
    return DFrame.read_from_kaggle('yasserh/advertising-sales-dataset', 
                             index_col=0, header=0, 
                             names=['', 'TV', 'Radio', 'Newspaper', 'Sales'])

def titanic():
    return DFrame.read_from_kaggle('titanic', 'train.csv', 'test.csv')

def diamonds():
    return DFrame.read_from_kaggle('shivam2503/diamonds')

def indian_liver():
    return DFrame.read_from_kaggle('uciml/indian-liver-patient-records')

def ames_housing():
    return DFrame.read_from_kaggle('house-prices-advanced-regression-techniques', 'train.csv', 'test.csv')
    
def air_passengers():
    return DFrame.read_from_kaggle('rakannimer/air-passengers')

def rossmann_store_sales():
    return DFrame.read_from_kaggle('rossmann-store-sales', 'train.csv', 'test.csv')

def california():
    return DFrame.read_from_kaggle('camnugent/california-housing-prices')
    
def heart_disease(**kwargs):
    return DFrame.read_from_kaggle('johnsmith88/heart-disease-dataset')
    
def speeddate(**kwargs):
    return DFrame.read_from_kaggle('annavictoria/speed-dating-experiment')
    
def occupancy():
    return DFrame.read_from_kaggle('robmarkcole/occupancy-detection-data-set-uci', 'datatraining.txt', 'datatest.txt')

def ag_news(language='basic_english', min_freq=1, collate='pad'):
    from torchtext.datasets import AG_NEWS
    train_iter, test_iter = read_torchtext(AG_NEWS)
    tc = TextCollection.from_iter(train_iter, test_iter, None, min_freq=min_freq).collate(collate)
    return tc

def bbc_news(language='basic_english', min_freq=1, collate='pad'):
    return TextCollection.from_csv('/data/datasets/bbc-text.csv', language=language, min_freq=min_freq).collate(collate)



from .dframe import DFrame, GroupedDFrame, to_numpy, show_warning
from .databunch import Databunch
from .textcollection import TextCollection
from .datasets import read_torchtext, wine_quality, telco_churn, movielens_ratings, movielens_movies, movielens_users, dam_outflow, boston_housing_prices, iris, bank_marketing, auto_mpg, big_mart_sales, advertising_sales, titanic, indian_liver, air_passengers, ames_housing, diamonds, california, occupancy, ag_news, ames_housing, boston_housing_prices, heart_disease, hotel_booking, hotel_test, hotel_test_score, bbc_news, speeddate, realestate
from .imagedframe import ImageDFrame, ImageDatabunch
from .imagedatasets import mnist, mnist3, crawl_images, filter_images, image_folder, create_path, cifar, fashionmnist
from .helper import read_excel, path_user, path_shared, read_from_function, read_from_package
from .kagglereader import create_kaggle_authentication, Kaggle
from .transformabledataset import TransformationXY, TransformableDataset
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt




import torch
import torch.nn as nn
import torchvision.models as models
from torch.hub import get_dir
import os

def last_container(last):
    try:
        l = last_container(last.children())
        if l is not None:
            return l
    except: pass
    try:
        if len(last._modules) > 0 and next(reversed(last._modules.values())).out_features > 0:
            return last
    except: pass
    
class DLModel(nn.Module):
    """
    A general purpose Deep model that will split an existing model between a base (lower layers) 
    and a final layer. This class provides methods to:
    - split_final: split off the final layer from a model
    - replace_final: replace the final layer with a Linear and a given number of output_nodes
    - freeze: freeze the base so that the weights are not trained
    - unfreeze: unfreeze the base, so the weights are trained
    
    Typical use for this class is for Transfer Learning.
    """
    def __init__(self, model):
        super().__init__()
        self.base, self.final = self.split_final(model)

    def split_final(self, model):
        container = last_container(model)
        name, last = container._modules.popitem()
        container.add_module(name, nn.Sequential())
        return model, last
    
    def replace_final(self, out_features):
        self.final = nn.Linear(self.final.in_features, out_features)

    def freeze(self):
        for c in list(self.base.children()):
            for p in c.parameters():
                p.requires_grad=False

    def unfreeze(self):
        for c in list(self.base.children()):
            for p in c.parameters():
                p.requires_grad=True

class Transfer(DLModel):
    """
    Create a model based on a (pretrained) torchvision model. The last layer is cut from the model
    and replaced with a linear that maps towards the number of given output_nodes. This model
    supports two-phase transfer learning through the freeze() and unfreeze() methods that
    allow to (un)freeze the model except the last layer.
    
    Args:
        out_features: int (None)
            if not None, the last layer in the model is replaced by a linear layer that maps
            to the given number of output nodes.
            
        model: tochvision.models.* (resnet34)
            the model to transfer. This class was only tested with models from torchvision.models
            but possibly works with other models as well.
            
        weights: str/bool (True)
            is passed to the torchvision model function, when True the model is downloaded from
            a PyTorch repository. The torchvision models are usually trained on ImageNet.
            In the TorchVision library, pretrained was replaced with weights. Passing
            None will not use pretrained weights, while passing 'DEFAULT' or the name of a
            set of pretrained weights will download these. In this Transfer model,
            True will be replaced by 'DEFAULT' and False by None.
            
        user: bool (False)
            the downloaded model is cached in an indicated location. By default, a shared folder
            ~/.pipetorch to save storage on a server with multiple users.
            Alternatively, when user=True the model is stored in a user specific ~/.pipetorchuser 
            folder. In case the users are not allowed write access on the shared folder.
    """
    
    def __init__(self, out_features=None, model=models.resnet34, weights=True, user=False, output_nodes=None):
        os.environ['TORCH_HOME'] = '~/.pipetorchuser' if user else '~/.pipetorch'
        if weights == True:
            weights = 'DEFAULT'
        elif weights == False:
            weights = None
        super().__init__(model(weights=weights))
        out_features = out_features or output_nodes
        if out_features is not None:
            self.replace_final(out_features)

    def forward(self, *X):
        h = self.base( *X )
        return self.final( h )




import nbformat
from pathlib import PosixPath
from nbconvert.preprocessors import ExecutePreprocessor
import os
import copy

class notebook:
    def __init__(self, path, file):
        self.file = file
        self.path = path

    @property
    def notebook(self):
        try:
            return self._notebook
        except:
            with open(self.fullpath) as fin:
                self._notebook = nbformat.read(fin, as_version=4)
            return self._notebook
    
    @property
    def stripped_notebook(self):
        try:
            return self._stripped_notebook
        except:
            nb = copy.deepcopy(self.notebook)
            cells = nb['cells']
            for i in range(len(cells)-1, -1, -1):
                cell = cells[i]
                if cell['cell_type'] == 'code' and 'Jupyter.notebook.session.delete' in cell['source']:
                    cell['source'] = ""
            self._stripped_notebook = nb
            return nb
    
    @property
    def fullpath(self):
        return self.path / self.file
    
    @property
    def runpath(self):
        return self.path.runfolder / self.file

    @property
    def errorpath(self):
        return self.path.errorfolder / self.file
    
    def exists(self):
        return self.fullpath.exists()
    
    def mtime(self):
        return os.path.getmtime(self.fullpath)
    
    def newer(self, other):
        return self.mtime() > other.mtime()
    
    def write_run(self):
        with open(self.runpath, 'w', encoding='utf-8') as fout:
            nbformat.write(self.stripped_notebook, fout)
        
    def write_error(self):
        with open(self.errorpath, 'w', encoding='utf-8') as fout:
            nbformat.write(self.stripped_notebook, fout)

    def execute(self):
        kernel = self.path.kernel or self.notebook.metadata.kernelspec.name
        ep = ExecutePreprocessor(timeout=self.path.timeout, 
                                 kernel_name=kernel)
        try:
            ep.preprocess(self.stripped_notebook, 
                          {'metadata': {'path': str(self.path)}})
            self.write_run()
            print('success')
        except:
            self.write_error()
            print('error')
         
        
class notebookpath(PosixPath):
    """
    A path containing notebooks, use execute_all to execute all notebooks in the folder and 
    store the finished notebooks in a subfolder
    """
    def __new__(cls, *path, **kwargs):
        return PosixPath.__new__(cls, *path)
        
    def __init__(self, *path, kernel=None,
                 runfolder='run', 
                 errorfolder='error',
                 timeout=600):
        super().__init__()
        self.kernel = kernel
        self._runfolder = runfolder
        self._errorfolder = errorfolder
        self.timeout = timeout
 
    def _subfolder(self, folder):
        r = self / folder
        r.mkdir(exist_ok=True)
        return r

    @property
    def runfolder(self):
        return self._subfolder(self._runfolder)

    @property
    def errorfolder(self):
        return self._subfolder(self._errorfolder)
        
    def notebook(self, file, path=None):
        path = path or self
        return notebook(path, file)
        
    def execute_all(self):
        for f in sorted(self.glob('*.ipynb')):
            print(f)
            nb = self.notebook(f.name)
            run = self.notebook(f.name, self.runfolder)
            error = self.notebook(f.name, self.errorfolder)
            
            if error.exists() and nb.newer(error):
                error.fullpath.unlink() 
            if run.exists() and nb.newer(run):
                run.fullpath.unlink()

            if not (run.exists() or error.exists()):
                nb.execute()



import torch
from torch import nn
import torchvision.models as models
from sklearn.metrics import *
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from .data.dframe import DFrame
from .data.databunch import Databunch
from .data.textcollection import TextCollection
from .data.datasets import read_torchtext, wine_quality, telco_churn, movielens_ratings, movielens_movies, movielens_users, dam_outflow, boston_housing_prices, iris, bank_marketing, auto_mpg, big_mart_sales, advertising_sales, titanic, indian_liver, air_passengers, ames_housing, diamonds, california, occupancy, ag_news, ames_housing, boston_housing_prices, heart_disease, hotel_booking, hotel_test, hotel_test_score, bbc_news, speeddate
from .data.imagedframe import ImageDFrame, ImageDatabunch
from .data.imagedatasets import mnist, mnist3, crawl_images, filter_images, image_folder, create_path, cifar, fashionmnist
from .data.kagglereader import Kaggle, create_kaggle_authentication
from .train.trainer import Trainer, Trainer as trainer
from .model.perceptron import Perceptron
from .model.convnet import ConvNet
from .model.transfer import Transfer, models

#from .data import *
#from .evaluate import *
#from .model import *
#from .train import *
#from .gpu import *
from .version import __version__
import warnings

class catch_warnings:
    def __enter__(self):
        warnings.filterwarnings("error")
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        warnings.filterwarnings("default")

def list_all(s):
    try:
        return s.__all__
    except:
        return [ o for o in dir(s) if not o.startswith('_') ]

#subpackages = [ jtorch.train, jtorch.train_modules, jtorch.train_diagnostics, jtorch.train_metrics, jtorch.jcollections ]

#subpackages = [ trainer ]

#__all__ = [ f for s in subpackages for f in list_all(s) ]





from .imagedframe import ImageDatabunch
import torch
import torch.nn.functional as F
from torchvision.datasets import MNIST, CIFAR10, FashionMNIST
from torchvision.transforms import transforms
from pathlib import Path
from getpass import getuser
import pandas as pd
import numpy as np
import os
from tqdm.notebook import tqdm
import io
from PIL import Image, ImageStat
from ..helper import is_documented_by
    
def create_path(p, mode=0o777):
    path = Path(p)
    os.makedirs(path, mode, exist_ok=True)
    return path

def image_folder():
    return f'/tmp/{getuser()}/images'

def _fast_torchvision(func, root=None, train=False, transform=None, testtransform=None, 
                      size=None, download=False, **kwargs):
    """
    A Dataset for MNIST, that caches the dataset on gpu for faster processing. 
    The images dataset are shaped SIZE.
    
    Arguments:
        root: str (None)
            local path where the dataset is stored. If None, first ~/.pipetorchuser is checked, then
            ~/.pipetorch (for shared datasets).
        
        train: bool (True)
            whether the train or test set is retrieved
            
        transform: callable (None)
            an optional function to perform additional transformations on the data
            not that the images are already tensors
            
        size: int (None)
            the required width/height of the image in pixels
            
        normalize: bool (True)
            Whether or not to normalize the images, which you normally should do, 
            but when you wish to inspect some of the images, you can turn it off.
            
        **kwargs: dict
            passed to torchvision's MNIST
    """
    if root is None:
        try:
            return func(root='~/.pipetorchuser', download=False, train=train, 
                        transform=transform, testtransform=testtransform, size=size, **kwargs)
        except: pass
        try:
            return func(root='~/.pipetorch', download=False, train=train, 
                        transform=transform, testtransform=testtransform, size=size, **kwargs)
        except: pass
        root = '~/.pipetorchuser'
    return func(root=root, download=download, train=train, 
                transform=transform, testtransform=testtransform, size=size, **kwargs)

class _FastCIFAR(CIFAR10):
    def __init__(self, root=None, train=True, transform=None, testtransform=None, 
                 normalize=True, size=None, **kwargs):
        super().__init__(root=root, train=train, **kwargs)
        self.transform=self.extend_transform(transform if train else testtransform, size)
        # Scale data to [0,1]
        self.data = torch.tensor(self.data).float().div(255)
        self.data = self.data.permute(0, 3, 1, 2)
        # Normalize it with the usual MNIST mean and std
        if normalize:
            self.normalized_mean = np.array([0.485, 0.456, 0.406 ])
            self.normalized_std = np.array([0.229, 0.224, 0.225 ])
            self.data[:,0].sub_(self.normalized_mean[0]).div_(self.normalized_std[0])
            self.data[:,1].sub_(self.normalized_mean[1]).div_(self.normalized_std[1])
            self.data[:,2].sub_(self.normalized_mean[2]).div_(self.normalized_std[2])
        self.targets = torch.tensor(self.targets)
        
    def extend_transform(self, transform, size=None):
        if size is None:
            return transform
        if transform is None:
            return transforms.Resize(size, antialias=True)
        return transforms.Compose([transforms.Resize(size, antialias=True), transform])
        
    def __getitem__(self, index):
        if self.transform:
            return self.transform(self.data[index]), self.targets[index]
        return self.data[index], self.targets[index]

class _FastFashionMNIST(FashionMNIST):
    def __init__(self, root=None, train=True, transform=None, testtransform=None, 
                 size=None, normalize=True, **kwargs):
        super().__init__(root=root, train=train, **kwargs)
        self.transform=self.extend_transform(transform if train else testtransform, size)
        self.data = self.data.unsqueeze_(dim=1).float().div(255)
        if normalize:
            self.data = self.data.div(0.5).sub(0.5)
            self.normalized_mean = np.array([0.5])
            self.normalized_std = np.array([0.5])

    def extend_transform(self, transform, size=None):
        if size is None:
            return transform
        if transform is None:
            return transforms.Resize(size, antialias=True)
        return transforms.Compose([transforms.Resize(size, antialias=True), transform])
        
    def __getitem__(self, index):
        if self.transform:
            return self.transform(self.data[index]), self.targets[index]
        return self.data[index], self.targets[index]

class _FastMNIST(MNIST):
    def __init__(self, root, train=True, transform=None, testtransform=None,
                 size=None, normalize=True, **kwargs):
        super().__init__(root=root, train=train, **kwargs)
        self.transform=self.extend_transform(transform if train else testtransform, size)

        # Scale data to [0,1]
        self.data = self.data.unsqueeze(1).float().div(255)
        if normalize:
            # Normalize it with the usual MNIST mean and std
            self.data.sub_(0.1307).div_(0.3081)
            self.normalized_mean = np.array([0.1307])
            self.normalized_std = np.array([0.3081])

    def extend_transform(self, transform, size=None):
        if size is None:
            return transform
        if transform is None:
            return transforms.Resize(size, antialias=True)
        return transforms.Compose([transforms.Resize(size, antialias=True), transform])
        
    def __getitem__(self, index):
        if self.transform:
            return self.transform(self.data[index]), self.targets[index]
        return self.data[index], self.targets[index]

class _FastMNIST3(_FastMNIST):
    def __init__(self, root, train=True, transform=None,
                 normalize=True, **kwargs):
        super().__init__(root, train=train, transform=transform, normalize=normalize, **kwargs)
        if normalize:
            self.normalized_mean = np.array([0.1307, 0.1307, 0.1307])
            self.normalized_std = np.array([0.3081, 0.3081, 0.3081, ])
        
    def __getitem__(self, index):
        img, target = self.data[index], self.targets[index]
        if self.transform:
            img = self.transform(img)
        img = torch.cat([img, img, img], axis=0)
        return img, target
    
@is_documented_by(_fast_torchvision, {'MNIST':'CIFAR10', 'SIZE':'(3,32,32)'})
def FastCIFAR(root=None, train=False, transform=None, testtransform=None, 
              normalize=True, download=False, **kwargs):
    return _fast_torchvision(_FastCIFAR, root=root, download=download, train=train, transform=transform, 
                             testtransform=testtransform, normalize=normalize, **kwargs)

@is_documented_by(_fast_torchvision, {'SIZE':'(1,28,28)'})
def FastMNIST(root=None, train=True, transform=None, testtransform=None,
              size=None, normalize=True, download=False, **kwargs):
    return _fast_torchvision(_FastMNIST, root=root, download=download, train=train, transform=transform, 
                             testtransform=testtransform, normalize=normalize, **kwargs)
                 
@is_documented_by(_fast_torchvision, {'MNIST':'FashionMNIST', 'SIZE':'(3,28,28)'})
def FastFashionMNIST(root=None, train=True, transform=None, testtransform=None,
                     size=None, download=False, normalize=True, **kwargs):
    return _fast_torchvision(_FastFashionMNIST, root=root, download=download, train=train, transform=transform,
                             testtransform=testtransform, normalize=normalize, **kwargs)
                 
@is_documented_by(_fast_torchvision, {'SIZE':'(3,28,28)'})
def FastMNIST3(root=None, train=True, transform=None, testtransform=None,
               size=None, normalize=True, download=False, **kwargs):
    return _fast_torchvision(_FastMNIST3, root=root, download=download, train=train, transform=transform, 
                             testtransform=testtransform, normalize=normalize, **kwargs)

def _fast_databunch(func, root=None, test=False, valid_perc=0.1, batch_size=64, test_batch_size=None,
                    transform=None, testtransform=None, size=None, num_workers=2, shuffle=True, 
                    pin_memory=False, balance=False, collate=None, classes=None, 
                    normalized_mean=None, normalized_std=None, **kwargs):
    '''
    An image_databunch of the MNIST dataset, shape is SIZE.
    
    Args:
        test: bool (False)
            if True, the test set is also included in the Databunch
            
        valid_perc: float (0.1)
            the size of the train set that is used for validation
    
        root=None, transform=None, size=None, normalize=True, **kwargs
            see FastMNIST for documentation

        batch_size=64, num_workers=2
            see ImageDatabunch for documentation
            
        **kwargs: dict (None)
            named argument that are passed to the TorchVision function that loads this dataset.
        
    Returns: ImageDatabunch
        An extension to a Databunch that provides additional support for images, like image normalization,
        transformation and show_batch to show an example.    
    '''
    train_ds = func(root, transform=transform, testtransform=testtransform, train=True, size=size, **kwargs)
    if test:
        test_ds = func(root, transform=transform, testtransform=testtransform, train=False, size=size, **kwargs)
    else:
        test_ds = None
    return ImageDatabunch.from_train_test_ds(train_ds, test_ds, valid_perc=valid_perc, batch_size=batch_size, 
                                             test_batch_size=test_batch_size, num_workers=num_workers, 
                                             shuffle=shuffle, pin_memory=pin_memory, balance=balance,
                                             collate=collate, classes=classes, normalized_mean=normalized_mean,
                                             normalized_std=normalized_std)

@is_documented_by(_fast_databunch, {'SIZE':'(1,28,28)'}) 
def mnist(root=None, test=False, valid_perc=0.1, batch_size=64, transform=None, testtransform=None, 
          size=None, normalize=True, num_workers=2, **kwargs):
    return _fast_databunch(FastMNIST, root=root, test=test, valid_perc=valid_perc, batch_size=batch_size, 
                           transform=transform, testtransform=testtransform, size=size, 
                           normalize=normalize, num_workers=num_workers, **kwargs)

@is_documented_by(_fast_databunch, {'SIZE':'(3,28,28)'}) 
def mnist3(root=None, test=False, valid_perc=0.1, batch_size=64, transform=None, testtransform=None, 
          size=None, normalize=True, num_workers=2, **kwargs):
    return _fast_databunch(FastMNIST3, root=root, test=test, valid_perc=valid_perc, batch_size=batch_size, 
                           transform=transform, testtransform=testtransform, size=size, 
                           normalize=normalize, num_workers=num_workers, **kwargs)

@is_documented_by(_fast_databunch, {'MNIST':'CIFAR10','SIZE':'(3,28,28)'}) 
def cifar(root=None, test=False, valid_perc=0.1, batch_size=64, transform=None, testtransform=None, 
          size=None, normalize=True, num_workers=2, **kwargs):
    return _fast_databunch(FastCIFAR, root=root, test=test, valid_perc=valid_perc, batch_size=batch_size, 
                           transform=transform, testtransform=testtransform, size=size, 
                           normalize=normalize, num_workers=num_workers, **kwargs)

@is_documented_by(_fast_databunch, {'MNIST':'FashionMNIST','SIZE':'(3,28,28)'}) 
def fashionmnist(root=None, test=False, valid_perc=0.1, batch_size=64, transform=None, testtransform=None, 
          size=None, normalize=True, num_workers=2, **kwargs):
    return _fast_databunch(FastFashionMNIST, root=root, test=test, valid_perc=valid_perc, batch_size=batch_size, 
                           transform=transform, testtransform=testtransform, size=size, 
                           normalize=normalize, num_workers=num_workers, **kwargs)

def _gis_args(keywords, output_directory=None, 
                 image_directory=None, limit=200, format='jpg', color_type='full-color', 
                 size='medium', type='photo', delay=0, **kwargs):
    """
    helper function for crawl_images, which uses google image search
    """
    if output_directory is None:
        output_directory = str(create_path(image_folder()))
    if image_directory is None:
        image_directory = '_'.join(keywords.split())
    arguments = {"keywords":keywords, 
                 "limit":limit, "format":format, "color_type":color_type, "size":size, "type":type, 
                 "delay":delay, "image_directory":image_directory, 
                 "output_directory":output_directory, "chromedriver":"/usr/bin/chromedriver" }
    arguments.update(kwargs)
    return arguments

def crawl_images(keywords, output_directory=None, 
                 image_directory=None, limit=200, format='jpg', color_type='full-color', 
                 size='medium', type='photo', delay=0, **kwargs):
    """
    Downloads images through Google Image Search, 
    see https://google-images-download.readthedocs.io/en/latest/arguments.html 
    for info on the arguments.
    The supporting library frequently breaks, so do not expect this to work.
    
    Arguments:
        keywords: the keywords passed to google image search to retrieve images
        limit: maximum number of images to retrieve (default=200). You will actually receive less iamges because many links will not work
        output_directory: base folder for the downloads (default: /tmp/username/images/)
        image_directory: subpath to store the images for this query (by default uses the query name)
        format: compression type of photos that are downloaded (default='jpg')
        color-type: default='full-color', see https://google-images-download.readthedocs.io/en/latest/arguments.html
        size: default='medium', see https://google-images-download.readthedocs.io/en/latest/arguments.html
        type: default='photo', see https://google-images-download.readthedocs.io/en/latest/arguments.html
        delay: default=0, to pause between downloads, see https://google-images-download.readthedocs.io/en/latest/arguments.html
        **kwargs: any additional arguments that google-images-download accepts.
    """
    try:
        from .google_images_download import googleimagesdownload
    except:
        raise NotImplemented('Need google images download for this')
    kwargs = _gis_args(keywords, output_directory=output_directory, image_directory=image_directory, 
             limit=limit, format=format, color_type=color_type, size=size, type=type, delay=delay, 
             **kwargs)
    response = googleimagesdownload()   #class instantiation
    paths = response.download(kwargs)   #passing the arguments to the function
    
def filter_images(keywords, folder=None, columns=4, height=200, width=200):
    """
    Removes duplicate images and shows the remaining images so that the user can manually select
    images to remove from the folder by pressing the DELETE button below.
    
    Arguments:
        keywords: subfolder of 'folder' in which the images are stored
        folder: folder/output_directory where the crawled images are stored (e.g. /tmp/username/images)
        columns (4): number of images displayed per row
        height (200): height of the images in pixels
        width (200): width of the images in pixels
    """
    import ipywidgets as widgets

    def on_click(button):
        for r in rows:
            if type(r) is widgets.HBox:
                for c in r.children:
                    checkbox = c.children[1]
                    if checkbox.value:
                        print(checkbox.description_tooltip)
                        os.remove(checkbox.description_tooltip)

    if folder is None:
        folder = Path(image_folder())
    keywords = '_'.join(keywords.split())
    imagefiles = [f for f in folder.glob(keywords + '/*')]
    rows = []
    cols = []
    bymean = {}
    for i, imgfile in enumerate(tqdm(imagefiles)):
        row = i // columns
        col = i % columns
        img = Image.open(imgfile)
        m = hash(tuple(ImageStat.Stat(img).mean))
        buff = io.BytesIO()   
        img.save(buff, format='JPEG')
        if m in bymean:
            os.remove(imgfile)
        else:
            bymean[m] = imgfile

        image = widgets.Image( value=buff.getvalue(), width=width, height=height )
        button = widgets.Checkbox( description='Delete', description_tooltip = str(imgfile) )
        box = widgets.VBox([image, button])
        cols.append(box)
        if len(cols) == columns:
            rows.append(widgets.HBox(cols))
            cols = []
                 
    if len(cols) > 0:
        rows.append(widgets.HBox(cols))
    button = widgets.Button( description='Delete' )
    button.on_click(on_click)
    rows.append(button)
    return widgets.VBox(rows)        

_ptdatasetslist = [('FastMNist B/W', 'pt.mnist()', 'https://github.com/y0ast/pytorch-snippets/tree/main/fast_mnist'),
            ('FastMNist RGB', 'pt.mnist3()', 'http://yann.lecun.com/exdb/mnist/'),
            ('FastCifar10', 'pt.cifar()', 'https://www.cs.toronto.edu/~kriz/cifar.html')
           ]
datasets = pd.DataFrame(_ptdatasetslist, columns=['dataset', 'method', 'url'])



import numpy as np
import pandas as pd
import copy
import itertools
import matplotlib.pyplot as plt

class EvaluatorResults(pd.DataFrame):
    _metadata = ['evaluator', 'df']
    
    @classmethod
    def from_evaluator(cls, 
                       evaluator):
        r = cls()
        r.evaluator = evaluator
        r.df = evaluator.df
        return r
    
    @property
    def _constructor(self):
        return EvaluatorResults

    def _copy_meta(self, 
                   df):
        df = self._constructor(df)
        for m in self._metadata:
            setattr(df, m, getattr(self, m))
        return df
    
    def clone(self):
        return copy.copy(self)
    
    def __finalize__(self, 
                     other, 
                     method=None, 
                     **kwargs):
        for name in self._metadata:
            object.__setattr__(self, name, getattr(other, name, None))
        return self
    
    def __repr__(self):
        r = self.drop( columns = [ c for c in self.columns if c[0] == '_' ] )
        return super(EvaluatorResults, r).__repr__()
    
    def _add(self, 
             row):
        for c in row.columns:
            if c not in self.columns:
                self.insert(len(self.columns), c, np.NaN)
        r = pd.concat([self, row], sort=True, ignore_index=True)
        #r = self.append(row, sort=True, ignore_index=True)
        r.evaluator = self.evaluator
        return r

    @property
    def train(self):
        r = self.loc[self.phase == 'train'].copy()
        return r
    
    @property
    def valid(self):
        r = self.loc[self.phase == 'valid'].copy()
        return r
    
    @property
    def test(self):
        r = self.loc[self.phase == 'test'].copy()
        return r

    def _plot_predict(self, 
                      pltfunction, 
                      label, 
                      x=None, 
                      xlabel = None, 
                      ylabel = None, 
                      title=None, 
                      interpolate=0, 
                      df=None, 
                      **kwargs):
        marker = itertools.cycle((',', '+', '.', 'o', '*')) 
        for _, row in self.iterrows():
            y = row['_predict']
            try:
                l = label(row)
            except:
                l = row[label]
            self.evaluator._plot(pltfunction, x=x, y=y, xlabel=xlabel, ylabel=ylabel, title=title, marker=next(marker), interpolate=interpolate, df=df, label=l, **kwargs)
        plt.legend()
   
    def _select(self, 
                select=None):
        if select is None:
            s = self.results
        elif type(select) is pd.core.series.Series:
            s = self.results[select]
        elif type(select) is EvaluatorResults:
            s = select
        elif type(select) is str:
            s = self.results[self.results.phase == select]
        else:
            raise ValueError('Unknown type passed for select')
        return s
    
    def _plot(self, 
              pltfunction, 
              x, y=None, 
              xlabel = None, 
              ylabel = None, 
              title=None, 
              loc='best', 
              fig=None, 
              legendargs={}, 
              **kwargs):
        f = _figure(self, x=x, y=y, xlabel=xlabel, ylabel=ylabel, title=title, fig=fig)
        pltfunction(f.graph_x, f.graph_y, **kwargs)      
        if 'label' in kwargs:
            self.evaluator._legend(f.fig, loc, **legendargs)

    def line(self, 
             x, 
             y=None, 
             xlabel = None, 
             ylabel = None, 
             title=None, 
             fig=None, 
             legendargs={}, 
             **kwargs):
        fig = fig or plt
        self._plot(fig.plot, x, y=y, xlabel=xlabel, ylabel=ylabel, title=title, 
                   fig=fig, legendargs=legendargs, **kwargs)
    
    def scatter(self, 
                x, 
                y=None, 
                xlabel = None, 
                ylabel = None, 
                title=None, 
                fig=None, 
                legendargs={}, 
                **kwargs):
        fig = fig or plt
        self._plot(fig.scatter, x, y=y, xlabel=xlabel, ylabel=ylabel, 
                   title=title, fig=fig, legendargs=legendargs, **kwargs)
        
    def line_metric(self, 
                    x, 
                    series='phase', 
                    y=None, 
                    xlabel = None, 
                    ylabel = None, 
                    title=None, 
                    label_prefix='', 
                    fig=None,
                    legendargs={},
                    **kwargs):
        self.evaluator.line_metric(x, series=series, select=self, y=y, xlabel=xlabel, ylabel=ylabel, 
                                   title=title, label_prefix=label_prefix, fig=fig, legendargs=legendargs, **kwargs)

    def scatter_metric(self, 
                       x, 
                       series='phase', 
                       y=None, 
                       xlabel = None, 
                       ylabel = None, 
                       title=None, 
                       label_prefix='', 
                       fig=None,
                       legendargs={},
                       **kwargs):
        self.evaluator.scatter_metric(x, series=series, select=self, y=y, xlabel=xlabel, ylabel=ylabel, 
                                      title=title, label_prefix=label_prefix, fig=fig, legendargs=legendargs, **kwargs)

class _figure:
    def __init__(self, 
                 results, 
                 x=None, 
                 y=None, 
                 xlabel = None, 
                 ylabel = None, 
                 title = None, 
                 fig=None ):
        fig = fig or plt
        self.evaluator = results.evaluator
        self.results = results
        self.x = x
        self.y = y
        self.xlabel = xlabel or self.x
        self.ylabel = ylabel or self.y
        self.fig = fig
        if title is not None:
            try:
                fig.title(title)
            except:
                fig.suptitle(title)
        try:
            fig.ylabel(self.ylabel)
        except:
            fig.set_ylabel(self.ylabel)
        try:
            fig.xlabel(self.xlabel)
        except:
            fig.set_xlabel(self.xlabel)
        self.results = self.results.sort_values(by=self.x)
        gx = [ row[self.x] for _, row in self.results.iterrows() ]
        gy = [ row[self.y] for _, row in self.results.iterrows() ]

    @property
    def x(self):
        return self._x
    
    @x.setter
    def x(self, 
          value):
        self._x = value or self.df._columnx[0]
        
    @property
    def y(self):
        return self._y
    
    @y.setter
    def y(self, 
          value):
        self._y = value or self.evaluator.metrics[0].__name__
        
    @property
    def xlabel(self):
        return self._xlabel
    
    @xlabel.setter
    def xlabel(self, 
               value):
        self._xlabel = value or self.x

    @property
    def ylabel(self):
        return self._ylabel
    
    @ylabel.setter
    def ylabel(self, 
               value):
        self._ylabel = value or self.y

    @property
    def graph_x(self):
        return [ row[self.x] for _, row in self.results.iterrows() ]

    @property
    def graph_y(self):
        return [ row[self.y] for _, row in self.results.iterrows() ]



#from torchtext.data.utils import get_tokenizer
from collections import Counter
#from torchtext.vocab import Vocab, build_vocab_from_iterator, FastText, GloVe
from torch.utils.data import DataLoader, IterableDataset
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data.dataset import random_split
from .databunch import Databunch
from .kagglereader import Kaggle
from .helper import path_shared, path_user, read_torchtext
import numpy as np
import torch
import copy

class TextCollection:
    """
    Setup an in memory text collection
    
    Args:
        train: TextDataSet
        valid: TextDataSet (None)
        test: TextDataSet (None)
        language: str ('basic_english')
            the langage used by the tokenizer
        min_freq: int (1)
            the minimum document frequency for words to be added to the vocabulary
        vocab: (None)
            a vocabulary to use. If None, a new vocabulary will be made from the trainset
        labels: LabelSet (None)
            a set of labes to use. If None, a LabelSet is constructed from the train set
        specials: tuple of str ('<unk>', 'pad')
            special tokens that are added to the vocabulary. Most commonly, <unk> is used
            for words that are not in the vocabulary and <pad> for making data points
            the same size.
        collate: callable (None)
            a function to prepare a batch, for example by padding all sentences to equal
            length.
    """
    
    def __init__(self, train, test=None, valid=None, language='basic_english', min_freq=1, vocab=None, 
                 labels = None, specials=('<unk>', '<pad>'), collate=None):
        self._vocab = vocab
        self.test = test
        self.train = train
        self.valid = valid
        self.language = language
        self.min_freq = min_freq
        self.labels = labels
        self.specials = specials
        self.__collate = collate
    
    @classmethod
    def from_iter(cls, train_iter, test_iter=None, valid_iter=None, **kwargs):
        """
        Reads a TextCollection from iterators. The most common use is from TorchText
        DataSets.
        
        Arguments:
            train_iter: str
                iterator that is used as the train set, e.g. 'train.csv'
            test_iter: str (None)
                iterator that is used as the test set, e.g. 'test.csv'
            valid_iter: str (None)
                iterator that is used as the valid set, e.g. 'valid.csv'
            **kwargs: see the TextCollection constructor for additional arguments
                      such as language, min_freq, vocab, labels, special, collate 
        Returns: TextCollection
        """
        train = TextDataSet.from_iter(train_iter)
        test = None if test_iter is None else TextDataSet.from_iter(test_iter)
        valid = None if valid_iter is None else TextDataSet.from_iter(valid_iter)
        return cls(train, test=test, valid=valid, **kwargs)
    
    @classmethod
    def from_csv(cls, train_filename, test_filename=None, valid_filename=None, **kwargs):
        """
        Reads a TextCollection from csv files. 
        
        Typically, the files are assumed to be organized in lines with the format: label, text
        Although often these files have a .csv extension, the files are not true csv files.
        
        Example:
            read_from_csv('train.csv', test_filename='test.csv', min_freq=1)
            to combine a train and test set in a single TextCollection

        Arguments:
            train_filename: str
                the filename that is used as the train set, e.g. 'train.csv'
            test_filename: str (None)
                the filename that is used as the test set, e.g. 'test.csv'
            valid_filename: str (None)
                the filename that is used as the valid set, e.g. 'valid.csv'
            **kwargs: see the TextCollection constructor for additional arguments
                      such as language, min_freq, vocab, labels, special, collate 
        Returns: TextCollection
        """
        train = TextDataSet.from_csv(train_filename)
        test = None if test_filename is None else TextDataSet.from_csv(test_filename)
        valid = None if valid_filename is None else TextDataSet.from_csv(valid_filename)
        return cls(train, test=test, valid=valid, **kwargs)

    @classmethod
    def from_kaggle(cls, dataset, train=None, test=None, valid=None, shared=True, force=False, **kwargs):
        """
        Reads a TextCollection from a Kaggle dataset. The files are forwarded to from_csv().
        
        The downloaded dataset is automatically stored so that the next time
        it is read from file rather than downloaded. 
        The dataset is stored by default in a folder
        with the dataset name in `~/.pipetorchuser`. 

        If the dataset is not cached, this functions requires a valid .kaggle/kaggle.json file, that you can 
        create manually or with the function `create_kaggle_authentication()`.

        Note: there is a difference between a Kaggle dataset and a Kaggle competition. For the latter, 
        you have to use `read_from_kaggle_competition`.

        Example:
            read_from_kaggle('yufengdev/bbc-text-categorization')
                to read/download `https://www.kaggle.com/datasets/yufengdev/bbc-text-categorization`

        Arguments:
            dataset: str
                the username/dataset part of the kaggle url, e.g. uciml/autompg-dataset for 
            train: str (None)
                the filename that is used as the train set, e.g. 'train.csv'
            test: str (None)
                the filename that is used as the test set, e.g. 'test.csv'
            valid: str (None)
                the filename that is used as the valid set, e.g. 'valid.csv'
            shared: bool (False)
                save the dataset in ~/.pipetorch instead of ~/.pipetorchuser, allowing to share downloaded
                files between users.
            force: bool (False)
                when True, the dataset is always downloaded
            **kwargs: see the TextCollection constructor for additional arguments
                      such as language, min_freq, vocab, labels, special, collate 
        Returns: TextCollection
        """
        k = Kaggle(dataset, shared=shared)
        if force:
            k.remove_user()
        train = k.file(train)
        if test is not None:
            test = k.file(test)
        if valid is not None:
            valid = k.file(valid)
        return cls.from_csv(train, test_filename=test, valid_filename=valid, **kwargs)

    @classmethod
    def from_torchtext(cls, func, min_freq=1, collate='pad', **kwargs): 
        """
        Reads a TextCollection from a torchtext data function.
        
        Arguments:
            func: callable
                from torchtext.datasets, a a function that behaves similarly and returns
                a train and test iterator for the dataset. By default the datasets are
                cached in the .pipetorchuser folder or used from the .pipetorch folder
                if present.
            **kwargs: see the TextCollection constructor for additional arguments
                      such as language, min_freq, vocab, labels, special, collate 
        Returns: TextCollection
        """
        try:
            train_iter, test_iter = read_torchtext( func )
        except:
            raise ValueError(f'cannot convert given function {func} to an iterator. Is it a torchtext data function?')
        return cls.from_iter(train_iter=train_iter, test_iter=test_iter, min_freq=min_freq, collate=collate, **kwargs)
            
    @property
    def _collate(self):
        return self.__collate
    
    def collate(self, collate):
        """
        Return a textcollection for which the dataset is prepared as:
        'pad': the sentences ae padded with a '<pad>' token to have equal length
        'offset': a batch is a list of tokenids and a seperate list of offsets indicate where a new sentence starts.
        ...: a cusrom collation function
        
        There are several options in TorchText to train with either padded of offset datasets.
        """
        if collate == 'offset':
            return OffsetTextCollection(self.train, test=self.test, valid=self.valid, 
                    language=self.language, min_freq=self.min_freq, vocab=self._vocab, labels = self.labels, 
                    specials=self.specials)
        if collate == 'pad':
            return PaddedTextCollection(self.train, test=self.test, valid=self.valid,  
                    language=self.language, min_freq=self.min_freq, vocab=self._vocab, labels = self.labels, 
                    specials=self.specials)
        r = TextCollection(self.train, test=self.test, valid=self.valid, 
                    language=self.language, min_freq=self.min_freq, vocab=self._vocab, labels = self.labels, 
                    specials=self.specials)
        r.__collate = collate
        return r   
    
    def split(self, *perc, test_perc=None, valid_perc=None):
        """
        Creates out-of-sample test and valid sets.
        
        Args:
            perc, test_perc, valid_perc: float[0-1]
                fraction of the data that is used resp. for the test and valid set.
                When a fixed test set is used, the first positional argument is used as
                valid_perc, otherwise as test_perc.

        Returns: TextCollection
            a shallow copy of this TextCollection that is configured to 
            split the data in a train, test and/or valid set

        Note: you cannot resplit a textcollection that was already split, or if it has a fixed validation or test set.
        """
        assert test_perc is None or self.test is None, 'You cannot specify a test_perc if a fixed test set is given'
        assert self.valid is None, 'You cannot resplit a text collection that was already split or that has a fixed validation set'
        assert self.vocab_not_exists, 'You cannot split a text collection when a vocabulary is already built'
        assert len(perc) < 2 or test_perc is None or valid_perc is None, 'Too many percentages'
        assert len(perc) < 3, 'Too many percentages'
        if len(perc) > 0:
            if test_perc is None:
                if self.test:
                    valid_perc = perc[0]
                else:
                    test_perc = perc[0]
            else:
                valid_perc = perc[0]
            if len(perc) == 2:
                test_perc, valid_perc = 1
                
        r = copy.copy(self)
        test_count = 0 if test_perc is None else round(test_perc * len(r.train))
        valid_count = 0 if valid_perc is None else round(valid_perc * len(r.train))
        train_count = len(r.train) - test_count - valid_count
        #print(len(r.train), train_count, valid_count, test_count)
        #print(len(r.train), sum([train_count, valid_count, test_count]))
        r.train, r.test, r.valid = random_split(r.train, [train_count, test_count, valid_count])
        return r
        
    @property
    def language(self):
        return self._language
        
    @language.setter
    def language(self, value):
        assert self.vocab_not_exists, 'You cannot specify a language after the vocabulary was built'
        if value is not None:
            self._language = value
            try:
                del self._tokenizer
                del self._vocab
                del self._labels
            except: pass
        
    @property
    def min_freq(self):
        return self._min_freq
        
    @min_freq.setter
    def min_freq(self, value):
        assert self.vocab_not_exists, 'You cannot specify a min_freq after the vocabulary was built'
        if value is not None:
            self._min_freq = value
    
    @property
    def tokenizer(self):
        """
        The tokenizer used to transform a text into a list of words/tokens.
        """
        try:
            return self._tokenizer
        except:
            from torchtext.data.utils import get_tokenizer
            self._tokenizer = get_tokenizer(self.language)
            return self._tokenizer

    @tokenizer.setter
    def tokenizer(self, value):
        assert self.vocab_not_exists, 'You cannot change the tokenizer after the vocabulary was built'
        self._tokenizer = value
        
    @property
    def vocab_not_exists(self):
        return self._vocab is None
        
    @property
    def vocab(self):
        """
        A (generated) PyTorch Vocab that maps all tokens in the training set to numbers.
        """
        if self._vocab is None:
            self._build_vocab()
        return self._vocab
    
    @vocab.setter
    def vocab(self, value):
        self._vocab = value
    
    @property
    def labels(self):
        """
        Collection of groud truth labels for the TextCollection
        """
        if self._labels is None:
            self._build_vocab()
        return self._labels

    @labels.setter
    def labels(self, value):
        self._labels = value
            
    def decode_sentence_index(self, words, index):
        """
        returns a list of tokens for the text at index in words.
        """
        return self.decode_sentence(words[index])
    
    def decode_sentence(self, tokenids):
        # lookup_tokens on vocab transforms List(int) -> List(str)
        return self.vocab.lookup_tokens(list(tokenids))

    def encode_sentence(self, sentence):
        # lookup_indices on vocab transforms List(str) -> List(int)
        return self.vocab.lookup_indices(self.tokenizer(sentence))

    def token_iterator(self, dataset):
        """
        returns: iterator over lists of tokenids that represent the original sentences
        may be overriden by subclassing TextCollection
        """
        def yield_tokens(dataset):
            for _, text in dataset:
                yield self.tokenizer(text)
        return yield_tokens(dataset)
    
    def _build_vocab(self):
        from torchtext.vocab import build_vocab_from_iterator

        labels = Counter([ l for l, _ in self.train ])
        self._vocab = build_vocab_from_iterator(self.token_iterator(self.train), specials=self.specials, special_first=True, min_freq=self.min_freq)
        self._vocab.set_default_index(self._vocab['<unk>'])
        self._labels = LabelSet(labels)
        
    def to_databunch(self, batch_size=32, shuffle=True, balance=False, **kwargs):
        """
        Return: a databunch (object with a dataloader for the train, valid, (test) set)
        batch_size (None): the batch_size used by the DataLoader
        shuffle (True): whether the samples are shuffled between epochs
        balance (False): if the training set is to be balanced (only works for binary classification)
        **kwargs: any other named argument that Databunch accepts
        
        Note: creating a databunch does not modify the original TextCollection, instead a shallow copy is made
        and creating the databunch usually triggers split, balance and creation of the vocabulary to be triggered.
        You can access the shallow copy of the textcollection through db.textcollection and the generated 
        vocabulary through db.vocab.
        """
        r = copy.copy(self)
        db = Databunch(None, r.train, r.test, r.valid, batch_size=batch_size, test_batch_size=batch_size, shuffle=shuffle, balance=balance, collate=r._collate, **kwargs)
        db.textcollection = r
        db.vocab = r.vocab
        if r._using_pretrained():
            db.pretrained_embeddings = r.pretrained_embeddings
            db.load_pretrained_embeddings_in_layer = r.load_pretrained_embeddings_in_layer
        return db
    
    def GloVe(self, name='6B', cache='/data/datasets', max_vectors=None):
        """
        Setup the use of pretrained GloVe embeddings. There are 4 sets available for download:
        name (6B): 6B, 42B, 840B or twitter.27B
        cache: shared folder to store downloaded embeddings. Do not change unless you know what you are doing
        since embeddings take up a lot of storage.
        max_vectors: use only the indicated number of vectors to save RAM. Since the tokens are sorted on frequency,
        using only the n most frequently appearing tokens works well in most cases.
        """
        from torchtext.vocab import GloVe
        r = copy.copy(self)
        r._pretrained = GloVe
        r._pretrained_params = {'name':name, 'cache':cache, 'max_vectors':max_vectors}
        return r
    
    def _using_pretrained(self):
        try:
            self._pretrained
            return True
        except:
            return False
    
    def pretrained_embeddings(self, dim):
        assert self._using_pretrained(), 'Can only prepare a table of pretrained embeddings if you configure using GloVe'
        self._pretrained_params['dim'] = dim
        self._embeddings = self._pretrained(**self._pretrained_params).get_vecs_by_tokens(self.vocab.get_itos())
        return self._embeddings
    
    def load_pretrained_embeddings_in_layer(self, layer):
        dim = embedding_layer.embedding_dim
        embedding_layer.weight.data = self.pretrained_embeddings(dim)              
        
class PaddedTextCollection(TextCollection):
    def _collate(self, batch):
        label_list, text_list = [], []
        for (_label, _text) in batch:
            label_list.append(self.labels[_label])
            processed_text = torch.tensor(self.encode_sentence(_text), dtype=torch.int64)
            text_list.append(processed_text)
        label_list = torch.tensor(label_list, dtype=torch.int64)
        text_list = pad_sequence(text_list, batch_first=True, padding_value=self.vocab['<pad>'])
        return text_list, label_list

    def decode_sentence_index(self, words, index):
        return super().decode_sentence(words[index].squeeze())
        
    def decode_sentences(self, words):
        return [ self.decode_sentence(words, i) for i in range(len(words))]
    
class OffsetTextCollection(TextCollection):
    def _collate(self, batch):
        label_list, text_list, offsets = [], [], [0]
        for (_label, _text) in batch:
            label_list.append(self.labels[_label])
            processed_text = torch.tensor(self.encode_sentence(_text), dtype=torch.int64)
            text_list.append(processed_text)
            offsets.append(processed_text.size(0))
        label_list = torch.tensor(label_list, dtype=torch.int64)
        offsets = torch.tensor(offsets[:-1]).cumsum(dim=0)
        text_list = torch.cat(text_list)
        return text_list, offsets, label_list

    def decode_sentence_index(self, words, offsets, index):
        start = offsets[index]
        end = offsets[index+1] if index < len(offsets)-1 else len(words)
        return super().decode_sentence(words[start:end])
        
    def decode_sentences(self, words, offsets):
        return [ self.decode_sentence(words, offsets, i) for i in range(len(offsets))]
    
class LabelSet:
    def __init__(self, labels):
        self.labels = labels
        self._itol = list(labels.keys())
        self._ltoi = {l:i for i, l in enumerate(self._itol)}
        
    def __getitem__(self, label):
        return self._ltoi[label]
    
    def __len__(self):
        return len(self._itol)
    
    def lookup_labels(self, labels):
        return [self._ltoi[l] for l in labels]
    
    def lookup_ints(self, ints):
        return [self._itol[i] for i in ints]

class TextDataSet:    
    @classmethod
    def from_iter(cls, it):
        r = cls()
        r.data = list(it)
        return r
        
    @classmethod
    def from_csv(cls, filename, header=True, delimiter=','):
        r = cls()
        with open(filename) as fin:
            r.data = [ l.split(delimiter, 1) for l in fin ]
        if header:
            r.data = r.data[1:]
        return r
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        return self.data.__getitem__(index)
